!
!   caltest_epn.f90  --- f90 program to test the CAL (EPN aspects only)
!
!   DESCRIPTION:
!       This program tests the correctness of the EPN-related CAL part by
!       performing a sequence of subroutine calls followed by
!       checks of the output. If any of these checks fails, the
!       program shall exit with a non-zero return value. If all tests
!       succeed 0 will be returned to the calling environment.
!
program caltest_epn

use cal
use errorhandling

implicit none
    logical                        :: isClose
    integer                        :: i, j, nx, ny
    integer(int16), dimension(:), pointer :: &
                                   rawX, rawY, badPixType, yextent
    logical, dimension(:), pointer :: uplinked
    integer(int16), dimension(0:5) :: &
                                   rawXtest = (/1, 23, 34, 56, 64, 33/),&
                                   rawYtest = (/1, 200, 111, 139, 32, 200/),&
                                   Ein = (/123, 456, 789, 999, 364, 888/), &
                                   rawXtest2, rawYtest2
    integer(int32), dimension(0:5) :: frame = (/1, 1, 1, 1, 1, 1/)

    real(single), dimension(:), pointer :: EinReal, EoutReal, EoutReal2
    integer(int16), dimension(:), pointer :: EoutInt, EoutInt2
    real(double), dimension(:), pointer :: xmilli, ymilli, zmilli
    integer(int16), dimension(:,:), pointer :: ibadpixmap

    type(BoresightT)                       :: boresight
    real(double), dimension(:,:), pointer  :: matrix
    real(single), dimension(:,:), pointer  :: PSFmap
    type(PsfT)                      :: psf
    type(PatternFractionServerT)    :: pfds;
    type(CurveT)                    :: curve
    type(PsfWindowT)                :: window
    real(double), dimension(:), pointer :: pattFracChan
    real(double)                    :: focalLength, plateScaleX, dum
    character(len=*), parameter     :: myBadPixFile = &
                                      './epn_badpix.ccf'
    type(CalInfo)                   :: badPixFileInfo
    integer(int32)                  :: instrumentOut, ccdChipIdOut, &
                                       accLevelOut
    real(single)                    :: sum

    call CAL_setState('epn.fits')
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      accuracyLevel = accLevelOut)
    if (instrumentOut /= EPN) &
        call fatal('caltest_epn', "instrumentOut /= EPN")
    if (ccdChipIdOut /= EPN_CCD_12) &
        call fatal('caltest_epn', "ccdChipIdOut /= EPN_CCD_12")

    call CAL_setState(EPN, ccdChipId=EPN_CCD_1, &
                      accuracyLevel = ACCURACY_LOW)
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      accuracyLevel = accLevelOut)
    call CAL_setState(temperatureCamera=299.d0)
    if (instrumentOut /= EPN) &
        call fatal('caltest_epn', "instrumentOut /= EPN")
    if (ccdChipIdOut /= EPN_CCD_1) &
        call fatal('caltest_epn', "ccdChipIdOut /= EPN_CCD_1")
    if (accLevelOut /= ACCURACY_LOW) &
        call fatal('caltest_epn', "accLevelOut /= ACCURACY_LOW")

    call CAL_ccfInfo('BADPIX', badPixFileInfo)
    call CAL_setCalibrationFile(myBadPixFile)

    !
    !   get bad pixel table / check a few
    !
    if (CAL_getBadPixelTableCode() /= INVALID_BAD_PIXEL_TABLE_CODE .or. &
        CAL_getAduconvCode() /= INVALID_BAD_PIXEL_TABLE_CODE) &
        call error('caltest_epn', &
                    "unexpected bad-pixel-table code")

    call CAL_setState(ccdChipId = EPN_CCD_1)
    call CAL_getBadPixelList(rawX, rawY, badPixType, yextent, uplinked)
    if (size(rawX) /= 20) &
        call fatal('caltest_epn', "size(rawX) /= 20")
    if (rawX(0) /= 3 .or. rawY(0) /= 174 .or. &
        yextent(0) /= 1 .or. badPixType(0) /= DEAD .or. &
        .not.uplinked(0)) &
        call fatal('caltest_epn', "something wrong with bad pixel 0 - node 0")
    if (rawX(19) /= 61 .or. rawY(19) /= 109 .or. &
        yextent(19) /= 2 .or. badPixType(19) /= DEAD .or. &
        uplinked(19)) &
        call fatal('caltest_epn', "something wrong with bad pixel 19")

    !   bad pixel map (ccd #1)
    call CAL_getBadPixelMap(ibadpixmap)
    do i=0, size(rawX)-1
        j = ibadpixmap(rawX(i)-1, rawY(i)-1)
        if ((mod(j, 10) /= badPixType(i))) then
            write(*, *) 'defect type of pixel', rawX(i), '/', rawY(i), &
                        ' is not right'
            call fatal('caltest_epn', "problem with bad pixel map")
        endif
        if (((j/10 .eq. 1) .and. .not.uplinked(i)) .or. &
            ((j/10 .eq. 0) .and. uplinked(i))) then
            write(*, *) 'uplink status of pixel', rawX(i), '/', rawY(i), &
                        ' is not right'
           call fatal('caltest_epn', "problem with uplink status of bad pixel")
        endif
    enddo
    call CAL_releaseMemory(rawX, rawY)
    call CAL_releaseMemory(badPixType, yextent)
    call CAL_releaseMemory(uplinked)
    call CAL_releaseMemory(ibadpixmap)

    !
    !   check get boresight matrix - sum of elements should be -1
    !
    !
    boresight = CAL_getBoresight(0.d0)
    matrix => CAL_toDirectionCosineMatrix(boresight)
    sum = abs(matrix(0, 0)) + abs(matrix(0, 1)) + abs(matrix(0, 2)) + &
          abs(matrix(1, 0)) + abs(matrix(1, 1)) + abs(matrix(1, 2)) + &
          abs(matrix(2, 0)) + abs(matrix(2, 1)) + abs(matrix(2, 2))
   write(*, *) sum
    if (.not. isClose(sum, 3.0023110)) &
        call fatal('caltest_epn', "boresight matrix is not right")

    !
    !   misc data
    !
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLength)
    call CAL_getMiscellaneousData('PLATE_SCALE_X', plateScaleX);
    if (.not. isClose(real(focalLength, single), 7493.2)) &
        call fatal('caltest_epn', "focal length is wrong")
    if (.not. isClose(real(plateScaleX, single), 4.128)) &
        call fatal('caltest_epn', "plate scale is wrong")

    !
    !   pixel to mm
    !
    call CAL_setState(randomize = .false.)
    call CAL_rawXY2mm(rawXtest, rawYtest, xmilli, ymilli, zmilli)
    call CAL_sacCoordToCamCoord2(xmilli, ymilli, zmilli, xmilli,ymilli,zmilli)
    call CAL_camCoord2ToCamCoord1(xmilli, ymilli, zmilli, xmilli,ymilli,zmilli)
    call CAL_camCoord1ToChipCoord(xmilli, ymilli, zmilli, rawX, rawY)
    do i=0, size(rawXtest)-1
       if (rawX(i) /= rawXtest(i) .or. &
           rawY(i) /= rawYtest(i)) &
        call fatal('caltest_epn', "XY<->mm not correct")
    enddo
    call CAL_releaseMemory(xmilli, ymilli, zmilli)
    call CAL_releaseMemory(rawX, rawY)

    !
    !   gain correction (low+high accuracy)
    !
    call CAL_setState(ccdChipId = EPN_CCD_1, randomize = .false., &
                      accuracyLevel = ACCURACY_LOW)

    call CAL_integerToReal(Ein, EinReal)
    call CAL_pnGainCorrect(rawXtest, rawYtest, EinReal, EoutReal)
    call CAL_realToInteger(EoutReal, EoutInt)
    write (*, *) EoutInt
    if (EoutInt(0) /= 119 .or. &
        EoutInt(1) /= 445 .or. &
        EoutInt(2) /= 775 .or. &
        EoutInt(3) /= 1000 .or. &
        EoutInt(4) /= 367 .or. &
        EoutInt(5) /= 863) &
        call fatal('caltest_epn', "low-acc. gain correction does not work")
    call CAL_releaseMemory(EinReal)
    call CAL_releaseMemory(EoutInt)


    call CAL_setState(ccdChipId = EPN_CCD_1, randomize = .false., &
                      accuracyLevel = ACCURACY_HIGH)

    call CAL_integerToReal(Ein, EinReal)
    call CAL_pnGainCorrect(rawXtest, rawYtest, EinReal, EoutReal)
    call CAL_realToInteger(EoutReal, EoutInt)
    write (*, *) EoutInt
    if (EoutInt(0) /= 121 .or. &
        EoutInt(1) /= 460 .or. &
        EoutInt(2) /= 801 .or. &
        EoutInt(3) /= 1038 .or. &
        EoutInt(4) /= 384 .or. &
        EoutInt(5) /= 896) &
        call fatal('caltest_epn', "high-acc. gain correction does not work")
    call CAL_releaseMemory(EinReal)
    call CAL_releaseMemory(EoutInt)

    !
    !   CTI correction
    !
    rawXtest2(0) = EPN_MIN_X
    rawXtest2(1) = EPN_MAX_X
    rawXtest2(2) = EPN_MAX_X
    rawXtest2(3) = EPN_MIN_X
    rawXtest2(4) = EPN_MIN_X
    rawXtest2(5) = EPN_MIN_X
    rawYtest2(0) = EPN_MIN_Y
    rawYtest2(1) = EPN_MIN_Y
    rawYtest2(2) = EPN_MAX_Y
    rawYtest2(3) = EPN_MAX_Y
    rawYtest2(4) = EPN_MIN_Y
    rawYtest2(5) = EPN_MAX_Y

    call CAL_setState(temperatureTracking = .true.)
    call CAL_pnCtiCorrect(rawXtest2, rawYtest2, frame, EoutReal, EoutReal2)
    call CAL_realToInteger(EoutReal2, EoutInt2)
    write (*, *) EoutInt2
    if (EoutInt2(0) /= 121 .or. &
        EoutInt2(1) /= 460 .or. &
        EoutInt2(2) /= 856 .or. &
        EoutInt2(3) /= 1105 .or. &
        EoutInt2(4) /= 384 .or. &
        EoutInt2(5) /= 957) &
        call fatal('caltest_epn', "CTI correction does not work")
    call CAL_releaseMemory(EoutInt2)
    call CAL_releaseMemory(EoutReal2)

    call CAL_setState(accuracyLevel = ACCURACY_LOW)

    !
    !   re-emission threshold
    !
    call CAL_setState(temperature = 165._double)
    write (*, *) CAL_pnReEmissionThreshold(1228.965_double, 150_int16);
    if (.not. isClose(real(CAL_pnReEmissionThreshold(1228.965_double, &
                          150_int16)), 26.499031842)) &
        call fatal('caltest_epn', "Re-emission threhold calc does not work")

    !
    !   get PSF map and check a window
    !
    psf = CAL_getPSF(1500._double, 0._double, 0._double);
    window%halfX = 1._double
    window%halfY = 1._double
    call CAL_psfImage(psf, window, PSFmap)

    nx = size(PSFmap, 1)/2
    ny = size(PSFmap, 2)/2
    write(*, *) PSFmap(nx, ny)
    write(*, *) PSFmap(nx+1, ny+1)
    write(*, *) PSFmap(nx+2, ny+2)
    write(*, *) PSFmap(nx+3, ny+3)
    write(*, *) PSFmap(nx+4, ny+4)

    if (.not. isClose(PSFmap(nx, ny), 7.3609924) .or. &
        .not. isClose(PSFmap(nx+1, ny+1), 6.6444502) .or. &
        .not. isClose(PSFmap(nx+2, ny+2), 5.1652832) .or. &
        .not. isClose(PSFmap(nx+3, ny+3), 3.4581327) .or. &
        .not. isClose(PSFmap(nx+4, ny+4), 1.9938920)) &
                              call fatal('caltest_epn', "PSF is wrong")
    call CAL_releaseMemory(PSFmap)


    !
    !   effective area
    !
    dum = CAL_getEffectiveArea(2000._double, 0._double, 0._double)
    if (.not. isClose(1.4253344950017295E+03, real(dum, single))) &
        call error('caltest_epn', "effective area is wrong")
    dum = CAL_getVignettingFactor(2000._double, 500._double, 0._double)
    write (*, *) 'vignetting at 2000 eV: ', dum
    if (.not. isClose(0.6119772367728963, real(dum, single))) &
        call error('caltest_epn', "vignetting is wrong")

    !
    !   pattern fraction data
    !
    pfds = CAL_getPatternFractionServer()
    call CAL_patternFractionAt(pfds, 20_int16, 100_int16)
    curve = CAL_getPatternFractionTotalQE(pfds)
    if (.not. isClose(0.9873661494985687, &
        real(CAL_curveValueAt(curve, 2000._double), single))) &
        call error('caltest_epn', "total QE is wrong")

    curve = CAL_getPatternFractionInEnergy(pfds, GRADE_DOUBLE)
    if (.not. isClose(0.3098717144763500, &
        real(CAL_curveValueAt(curve, 2000._double), single))) &
        call error('caltest_epn', "double pattern fraction is wrong")

    pattFracChan => CAL_getPatternFractionInChannel(pfds, GRADE_SINGLE)
    if (size(pattFracChan) /= 4096 .or. &
        .not. isClose(0.7919080257415771, real(pattFracChan(100), single))) &
        call error('caltest_epn', "single pattern fraction is wrong")

    call message(AppMsg, Verbose,"all tests passed")
end program


logical function isClose(value, trueval)
use types

    real(single), intent(in)    :: value, trueval
    real(single)                :: absval, abstrueval

    isClose = .true.
    if (value * trueval .lt. 0.) isClose = .false.
    absval = abs(value)
    abstrueval = abs(trueval)
    if (absval .lt. .99999 * abstrueval .or. &
        absval .gt. 1.0001 * abstrueval) isClose = .false.
end function
