!
!   caltest_rgs.f90  --- f90 program to test the CAL (RGS aspects only)
!
!   DESCRIPTION:
!       This program tests the correctness of the RGS-related CAL part by
!       performing a sequence of subroutine calls followed by
!       checks of the output. If any of these checks fails, the
!       program shall exit with a non-zero return value. If all tests
!       succeed 0 will be returned to the calling environment.
!
program caltest_rgs

use errorhandling
use types
use cal

implicit none
    logical                         :: isClose, isAlmostClose
    integer(int32)                  :: instrumentOut, ccdChipIdOut, &
                                       ccdNodeIdOut, accLevelOut, i, ocbOut, &
                                       n1, n2, n
    integer(int32), dimension(0:0) :: nmax
    integer(int16), dimension(0:5) :: &
                                   rawXtest = (/1, 23, 34, 56, 118, 567/),&
                                   rawYtest = (/1, 234, 43, 56, 98, 876/),&
                                   Ein = (/123, 456, 789, 999, 364, 888/), &
                                   EinOrg
    integer(int16), dimension(0:5) :: &
                                   chipXnom = (/1, 23, 34, 969, 907, 458/)
    integer(kind=int16), dimension(:), pointer      :: chipX, chipY
    integer(int8), dimension(0:5)  :: &
                                   nodes = (/0, 0, 0, 1, 1, 1/)
    integer(int8), dimension(0:5)  :: &
                                   shapes = (/1, 1, 1, 1, 1, 1/)
    integer(int16), dimension(0:5) :: &
                                   separation = (/0, 0, 0, 0, 0, 0/)
    integer(int16), dimension(:), pointer :: PHA, PI
    real(single), dimension(:), pointer :: EoutReal, PHAreal, PIreal,&
                                           accThresh, rejThresh
    real(double), dimension(0:5)  :: &
                                   timeTags = (/0._double, 0._double, &
                                                0._double, 0._double, &
                                                0._double, 0._double/)
    integer(int16)                       :: rawXdum, rawYdum
    real(double)                         :: rejThreshdum, maxvalue
    real(kind=double), dimension(:), pointer :: redist
    real(kind=double)                        :: xAnglesMin, binWidth, sum
    real(kind=double), dimension(0:799)      :: xAngles
    real(kind=double), dimension(:), pointer :: xpsf
    real(kind=double), dimension(:), pointer :: x, x2, y, y2, z, z2, beta, chi

    call CAL_setState('rgs.fits')
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      ccdNodeId=ccdNodeIdOut, ocb = ocbOut, &
                      accuracyLevel = accLevelOut)
    if (instrumentOut /= RGS2) &
        call error('caltest_rgs', "instrumentOut /= RGS2")
    if (ccdChipIdOut /= RGS_CCD_7) &
        call error('caltest_rgs', "ccdChipIdOut /= RGS_CCD_7")
    if (ccdNodeIdOut /= 0) &
        call error('caltest_rgs', "ccdNodeIdOut /= 0")
    if (ocbOut /= 5) &
        call error('caltest_rgs', "ocbOut /= 5")

    !
    !   PIXCOORD1<->CHIPCOORD<->CAMCOORD1<->CAMCOORD2<->SACCOORD<->ROWCOORD
    !   conversion
    !
    call CAL_setState(ocb=1)
    call CAL_pixCoord1ToChipCoord(rawXtest, rawYtest, chipX, chipY, nodes)
    call CAL_chipCoordToCamCoord1(chipX, chipY, x, y, z)
    call CAL_camCoord1ToCamCoord2(x, y, z, x, y, z)
    call CAL_camCoord2ToSacCoord(x, y, z, x, y, z)
    call CAL_SacCoordToRowCoord(x, y, z, beta, chi)
    call CAL_RowCoordToSacCoord(beta, chi, x2, y2, z2)
    do n=0, size(rawXtest)-1
         write(*, *) rawXtest(n), rawYtest(n), chipX(n), chipY(n)
         write(*, *) x(n), y(n), z(n), beta(n), chi(n)
         write(*, *) x(n), x2(n), y(n), y2(n), z(n), z2(n)
         if (rawYtest(n) /= chipY(n) .or. &
             chipX(n) /= chipXnom(n)) &
         call error('caltest_rgs', 'PIXCOORD1->CHIPCOORD conv. does not work')
         if (.not. isAlmostClose(x(n), x2(n)) .or. &
             .not. isAlmostClose(y(n), y2(n)) .or. &
             .not. isAlmostClose(z(n), z2(n))) &
         call error('caltest_rgs', 'ROWCOORD->SACCCORD conv. does not work')
    enddo

    !
    !   offset correction
    !
    call CAL_setCalibrationFile('rgsoffsetcalc.ds');
    EinOrg = Ein
    call CAL_setState(instrument=RGS1, ccdChipId = RGS_CCD_8, ccdNodeId = 1, &
                      randomize = .false., ocb = 1, &
                      observationStartTime=51859.7083333333358d0)
    call CAL_rgsOffsetCorrect(timeTags, rawXtest, rawYtest, shapes, nodes, &
                              Ein, EoutReal)
    call CAL_realToInteger(EoutReal, PHA)
    write(*, *) Ein
    write(*, *) PHA
    if (PHA(0) /= -310 .or. &
        PHA(1) /= 36 .or. &
        PHA(2) /= 384 .or. &
        PHA(3) /= 577 .or. &
        PHA(4) /= -61 .or. &
        PHA(5) /= 401) then
        call error('caltest_rgs', "Offset correction does not work")
    endif
    call CAL_setState(ccdChipId = RGS_CCD_1, ccdNodeId = 0)

    !
    !   event thresholds
    !
    call CAL_rgsGetEventThresholds(rejThresh, accThresh, nodes(5))
    write(*, *) 'rejection thresholds for CCD #1/node1: ', rejThresh
    write(*, *) 'acceptance thresholds for CCD #1/node1: ', accThresh

    !
    !   CTI correction
    !
    rawXtest(0) = RGS_MIN_X
    rawXtest(1) = RGS_MAX_X
    rawXtest(2) = RGS_MAX_X
    rawXtest(3) = RGS_MIN_X
    rawXtest(4) = RGS_MIN_X
    rawYtest(0) = RGS_MIN_Y
    rawYtest(1) = RGS_MIN_Y
    rawYtest(2) = RGS_MAX_Y
    rawYtest(3) = RGS_MAX_Y
    rawYtest(4) = RGS_MIN_Y
    call CAL_rgsCtiCorrect(timetags, rawXtest, rawYtest, nodes, shapes, &
                           separation, EoutReal, PHAreal)
    call CAL_realToInteger(PHAreal, PHA)
    write(*, *) PHA(0), PHA(1), PHA(2), PHA(3), PHA(4), PHA(5)
    if (PHA(0) /= -310 .or. &
        PHA(1) /= 37 .or. &
        PHA(2) /= 566 .or. &
        PHA(3) /= 823 .or. &
        PHA(4) /= -61 .or. &
        PHA(5) /= 423) &
        call error('caltest_rgs', "CTI correction does not work")

    !
    !   gain correction
    !
    call CAL_gainCorrect(timeTags, PHAreal, PIreal, nodes)
    call CAL_realToInteger(PIreal, PI)
    write(*, *) PI(0), PI(1), PI(2), PI(3), PI(4), PI(5)
    if (PI(0) /= -284 .or. &
        PI(1) /= 34 .or. &
        PI(2) /= 520 .or. &
        PI(3) /= 695 .or. &
        PI(4) /= -52 .or. &
        PI(5) /= 357) &
       call error('caltest_rgs', "gain correction does not work")

    call CAL_releaseMemory(PIreal, PHAreal)
    call CAL_releaseMemory(PHA, PI)


    !
    ! redistribution
    !
    rawXdum = 1
    rawYdum = 1
    rejThreshdum = 1.
    call CAL_rgsGetRedistribution(1000._double, 1_int16, 1_int16, &
                                  rejThreshdum, redist)
    write(*, *) 'size of redistribution spectrum', size(redist)
    if (size(redist) /= 4096) &
        call error('caltest_rgs', "incorrect size of redistribution spectrum")
    nmax = maxloc(redist)
    maxvalue = maxval(redist)
    n1 = 0
    n2 = 0
    do i=0, size(redist)-1
       if (redist(i) .ge. .5*maxvalue) then
           n1 = i
           exit
       endif
    enddo
    do i=n1+1, size(redist)-1
       if (redist(i) .le. .5*maxvalue) then
           n2 = i
           exit
       endif
    enddo

    write(*, *) n1, nmax(0), n2, n2-n1
    if (nmax(0) < 990 .or. nmax(0) .gt. 995) &
        call error('position of redistribution peak is wrong')
    if (n2-n1 .gt. 105 .or. n2-n1 .lt. 80) &
        call error('width of redistribution peak is wrong')

    !
    !   the same test for E=2000 eV
    !
    call CAL_rgsGetRedistribution(2000._double, 1_int16, 1_int16, &
                                  rejThreshdum, redist)
    write(*, *) 'size of redistribution spectrum', size(redist)
    if (size(redist) /= 4096) &
        call error('caltest_rgs', "incorrect size of redistribution spectrum")
    nmax = maxloc(redist)
    maxvalue = maxval(redist)
    n1 = 0
    n2 = 0
    do i=0, size(redist)-1
       if (redist(i) .ge. .5*maxvalue) then
           n1 = i
           exit
       endif
    enddo
    do i=n1+1, size(redist)-1
       if (redist(i) .le. .5*maxvalue) then
           n2 = i
           exit
       endif
    enddo

    write(*, *) n1, nmax(0), n2, n2-n1
    if (nmax(0) < 1980 .or. nmax(0) .gt. 1985) &
        call error('position of redistribution peak is wrong')
    if (n2-n1 .gt. 110 .or. n2-n1 .lt. 100) &
        call error('width of redistribution peak is wrong')


    !
    !   cross dispersion PSF
    !
    call CAL_setState(ocb = 1)
    xAnglesMin =  -.1 * acos(-1.)/180.
    binWidth = (-xAnglesMin - xAnglesMin)/800.;
    do i=0, 799
        xAngles(i) = xAnglesMin + (real(i)-.5) * binWidth
    enddo
    call CAL_rgsGetCrossPSF(0._double, .056335894_double, 0._double, &
                            xAngles, binWidth, xpsf)
    write(*, *) 'size of redistribution spectrum', size(xpsf)
    if (size(xpsf) /= 800) &
        call error('caltest_rgs', "incorrect size of cross-psf probability")
    sum = 0._double
    do i = 0, 799
        sum = sum + xpsf(i)
    enddo
    write(*, *) 'Sum of probabilities: ', sum
    if (.not. isClose(sum, 1._double)) &
        call error('caltest_rgs', "incorrect normalization of cross-psf");

    call message(AppMsg, Verbose, "all tests passed")
end program


logical function isAlmostClose(value, trueval)
use types
    real(double), intent(in)   :: value, trueval
    real(double)               :: absval, abstrueval

    isAlmostClose = .true.
    absval = abs(value)
    abstrueval = abs(trueval)

    if (absval .lt. .98 * abstrueval .or. &
        absval .gt. 1.02 * abstrueval) isAlmostClose = .false.
end function


logical function isClose(value, trueval)
use types

    real(double), intent(in)    :: value, trueval
    real(double)                :: absval, abstrueval

    isClose = .true.
    if (value * trueval .lt. 0.) isClose = .false.
    absval = abs(value)
    abstrueval = abs(trueval)

    if (absval .lt. .99 * abstrueval .or. &
        absval .gt. 1.001 * abstrueval) isClose = .false.
end function


