//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	BackgroundDataServer.h --- background related calibration entities
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2002-06-14: 1.0.0: - first version with support for RGS HTR mode
//
//	AUTHOR:
//		U. Lammers, Fri Jun 14 15:31:51 MEST 2002
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/BackgroundDataServer.h,v 1.1 2002/06/26 14:32:15 ulammers Exp $
//

#ifndef BACKGROUND_DATA_SERVER_H
#define BACKGROUND_DATA_SERVER_H

#include "Ccf.h"
#include "CalAtomsBase.h"

class BackgroundSpectrum : public std::vector<double> {
public:
	BackgroundSpectrum() : _nChan(0) {}

	void pushChannel(double value) { _channs.push_back(_nChan++);
									 push_back(value); }
	const std::vector<unsigned> &channels() const { return _channs; }

	~BackgroundSpectrum() {}

private:
	std::vector<unsigned> _channs;
	unsigned			  _nChan;
};

class HtrBackgroundIntensity;	//	forward declaration

//
//	---( BackgroundDataServer )------------------------------------------------
//
class BackgroundDataServer : public CalAtom,
		  public virtual CalChild<CalAtomCollection<BackgroundDataServer> > {
public:
	BackgroundDataServer(CalAtomCollection<BackgroundDataServer> *parent = 0)
		: CalChild<CalAtomCollection<BackgroundDataServer> >(parent) {}
		//
		//	PURPOSE:
		//		Construct BackgroundDataServer object
		//

	static unsigned id() { return Ccf::Background; }

	void dump(std::ostream &) const {}

	virtual BackgroundSpectrum particle(double time = 0, double r = 0.) const;
	virtual BackgroundSpectrum diffuseX(double time = 0, double r = 0.) const;
		//
		//	PURPOSE:
		//		Return spectrum representative for particle- and diffuse X-ray
		//		background
		//


	virtual HtrBackgroundIntensity betaSelect(double beta) const;
		//
		//	PURPOSE:
		//		Return HtrBackgroundIntensity object for given dispersion
		//		angle <beta>; applicable for RGS HTR mode only
		//
		//	ARGUMENTS:
		//		beta	: dispersion angle [rad]
		//


	virtual ~BackgroundDataServer() {}
		//
		//	PURPOSE:
		//		Destroy BackgroundDataServer object
		//
}; // Background


class EpicBackgroundDataServer : public BackgroundDataServer {
public:
	EpicBackgroundDataServer(const CcfConstituentTable &particle,
							 const CcfConstituentTable &diffuse,
							 const CcfConstituentTable &diffuseMod,
						 CalAtomCollection<BackgroundDataServer> *parent = 0);

	BackgroundSpectrum particle(double time = 0, double r = 0.) const;
	BackgroundSpectrum diffuseX(double time = 0, double r = 0.) const;
		//
		//	PURPOSE:
		//		See above
		//

	~EpicBackgroundDataServer();


private:
	double				_particleP0, _particleP1;
	BackgroundSpectrum	_diffuse;
	double				_diffModP0, _diffModP1, _diffModP2;
};


class HtrBackgroundIntensity {
public:
	HtrBackgroundIntensity(double beta) : _beta(beta) {}

	double accumulatedIntensity(double PIfrom, double PIto) const;
		//
		//	PURPOSE:
		//		Accumulate HTR background intensity maps for previously
		//		selected dispersion angle value in given PI interval
		//
		//	ARGUMENTS:
		//		PIfrom, PIto: PI range for the intensity accumulation
		//
		//	RETURNS:
		//		Integrated HRT background intensity in the PI interval
		//		[PIfrom, PIto]; accumulatedIntensity(PI, PI) == 0
		//


	double beta() const { return _beta; }
		//
		//	PURPOSE:
		//		Return beta value this object is valid for
		//

	~HtrBackgroundIntensity() {}

private:
	double	_beta;
};


class RgsBackgroundDataServer : public BackgroundDataServer {
public:
	RgsBackgroundDataServer(
						CalAtomCollection<BackgroundDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Construct RgsBackgroundDataServer object


	HtrBackgroundIntensity betaSelect(double beta) const;
		//
		//	PURPOSE:
		//		Construct HtrBackgroundIntensity object applicable for at
		//		given angle beta
		//
		//	ARGUMENTS:
		//		beta	: dispersion angle [rad]
		//


	~RgsBackgroundDataServer();
		//
		//	PURPOSE:
		//		Destroy RgsBackgroundDataServer object
		//
};

#endif

