//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Boresight --- the Boresight class
//
//	DESCRIPTION:
//		A class 'Boresight' is defined which provides access to
//		the boresight misalignment angles and the associated matrix
//		for a specific instrument. The instrument boresight frame
//		is defined with respect to the satellite frame (see [2]):
//			o The positive X-axis extents from the origin of the CAMCOORD2
//			  frame along the optical axis of the telescope into the sky.
//			o The +Z axis is in the +X/+Z_SC plane perpendicular to the
//			  +X axis
//			o The +Y axis completes the orthogonal, right-handed system
//
//	CHANGE HISTORY:
//		UL, 1999-08-17: 1.0.0: - first version from CalAtomsCommon
//		UL, 2001-03-02: 1.1.0: - added withSacCoordOrientation()
//							   - added projectOntoTelCoord()
//
//	AUTHOR:
//		U. Lammers (UL), Tue Aug 17 11:02:56 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Star Tracker/Instrument Alignment Calibration ICD,
//			 R. Mugellesi, et al., XMM-MOC-ICD-0025-OAD
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/Boresight.h,v 1.7 2002/05/24 10:59:16 gvacanti Exp $
//

#ifndef BORESIGHT_H
#define BORESIGHT_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "Celestial.h"

#include "GeometryDataServer.h"

//
//	---( Boresight )----------------------------------------
//
class Boresight : public CalAtom,
	 public virtual CalChild<CalAtomCollection<Boresight> > {
public:
	Boresight(const EulerAngles &bsAngles,
			  CalAtomCollection<Boresight> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate Boresight object from Euler angles
		//


	static unsigned id() { return Ccf::BoreSight; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual const EulerAngles &angles(double time = 0);
		//
		//	PURPOSE:
		//		Return the boresight misalignment at a given point in time in
		//		the form of three Euler angles
		//
		//	ARGUMENTS:
		//		time:	point in time in the TIMCOORD frame at which the
		//				boresight is requested
		//


	virtual void angles(const EulerAngles &eulerAngles);
		//
		//	PURPOSE:
		//		Set the boresight to a constant value defined by a given
		//		set of Euler angles; subsequent invocation of
		//		Boresight::angles(t) will yield this Euler angles for all
		//		values of <t>.
		//
		//	ARGUMENTS:
		//		eulerAngles: Euler angles defining instrument boresight
		//					 in the above described frame [2]
		//


	virtual const AttitudeMatrix &matrix(double time) const;
		//
		//	PURPOSE:
		//		Return the boresight misalignment at a given point in time in
		//		the form of a proper orthogonal rotation matrix A. The matrix
		//		can be used to map vectors from the S/C frame into the
		//		boresight frame:
		//			V_bs = A * V_sc
		//


	virtual const Boresight &withSacCoordOrientation(
											const SpacecraftAttitude &) const;
		//
		//	PURPOSE:
		//		Define an orientation of SACCOORD in equatorial J2000 reference
		//		frame
		//


	XmmTelescopePointList projectOntoTelCoord(const EquatorialDirection &)
	const;
	XmmTelescopePointList projectOntoTelCoord(
							const std::vector<EquatorialDirection> &) const;
		//
		//	PURPOSE:
		//		Project an equatorial direction/list of directions onto
		//		TELCOORD frame
		//


	virtual ~Boresight() {}
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members
	EulerAngles		_angles;
	AttitudeMatrix	_m;
	mutable double	_mIsValidForTime,
					_anglesIsValidForTime;

	mutable bool	_saccoordDefined;
	mutable AttitudeMatrix	_saccoordMatrix;
}; // Boresight

#endif
