//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalAtomServer.cc --- implementation of CalAtomCollectionServer
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1998-04-24: 1.0.0: - skeleton
//		UL, 1999-05-12: 1.1.0: - changed fatal -> error and changed all
//								 error tags
//
//	AUTHOR:
//		UL, Fri Apr 24 09:00:59 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/CalAtomServer.cc,v 1.38 2002/06/26 14:32:15 ulammers Exp $
//

#include "CalServer.h"
#include "CalAtomServer.h"

using namespace std;

CalAtomCollectionServer::CalAtomCollectionServer(const CalServer *server)
	: CalChild<CalServer>(server)
{
	assert(server);
	size_t nEntries = Xmm::N_INSTRUMENTS * Ccf::N_CALDATATYPES;
	CalAtomCollectionBase *null = 0;
	_atomsCollection = new vector<CalAtomCollectionBase *>(nEntries, null);
} // CalAtomCollectionServer::CalAtomCollectionServer()


void CalAtomCollectionServer::restart()
{
	unsigned N = _atomsCollection->size();
	for (unsigned n=0; n<N; n++) {
		delete (*_atomsCollection)[n]; (*_atomsCollection)[n] = 0;
	}
} // CalAtomCollectionServer::restart()


unsigned CalAtomCollectionServer::index(const Xmm::Instrument instrumentId,
										unsigned n) const
{
	return instrumentId * Ccf::N_CALDATATYPES + n;
} // CalAtomCollectionServer::index()


CalAtomCollectionBase *CalAtomCollectionServer::getEntry(Xmm::Instrument id,
														 unsigned n) const
{
	unsigned pos = index(id, n);

	if ((*_atomsCollection)[pos] &&
		(!parent()->state()->filter()->isActive() ||
			parent()->state()->filter()->get() ==
				(*_atomsCollection)[pos]->filter()) &&
		(*_atomsCollection)[pos]->accuracy() ==
				parent()->state()->accuracyLevel()->get())
			return (*_atomsCollection)[pos];

	delete (*_atomsCollection)[pos]; (*_atomsCollection)[pos] = 0;

	return static_cast<CalAtomCollectionBase *>(0);
} // CalAtomCollectionServer::getEntry()


void CalAtomCollectionServer::setEntry(Xmm::Instrument id, unsigned n,
									   CalAtomCollectionBase *p) const
{
	(*_atomsCollection)[index(id, n)] = p;
} // CalAtomCollectionServer::setEntry()


template <class AtomType>
CalAtomCollection<AtomType> *
CalAtomCollectionServer::getCollection(const XmmInstrument *instrument,
									   AtomType *) const
{
	CalAtomCollectionBase *ptr =
			getEntry(instrument->id(), AtomType::id());
	if (ptr)
		return dynamic_cast<CalAtomCollection<AtomType> *>(ptr);

	CalAtomCollection<AtomType> *p =
			new CalAtomCollection<AtomType>(
					instrument,
					parent()->ccf(),
					parent()->state()->filter()->isActive() ?
						parent()->state()->filter()->get() :
							Xmm::FILTER_NONE,
					parent()->state()->accuracyLevel()->get());

	setEntry(instrument->id(), AtomType::id(), p);

	return p;
}


#define getCol(a)														\
template CalAtomCollection<a> *CalAtomCollectionServer::getCollection(	\
					const XmmInstrument *instrument, a *) const;

getCol(AbsorptionDataServer)
getCol(BackgroundDataServer)
getCol(BadPixelDataServer)
getCol(Boresight)
getCol(CanonicalCrossPsf)
getCol(CanonicalRedist)
getCol(CcdQuantumDataServer)
getCol(GeometryDataServer)
getCol(GratingDataServer)
getCol(CtiCorrector)
getCol(DarkFrameDataServer)
getCol(EffectiveAreaDataServer)
getCol(EncircledEnergyDataServer)
getCol(EnergyCorrector)
getCol(FilterDataServer)
getCol(HKvalRangesServer)
getCol(MiscDataServer)
getCol(ModeParameters)
getCol(OmColorTransformator)
getCol(OmGeoLinearizer)
getCol(OmFlatFieldServer)
getCol(OmFluxConverter)
getCol(OmPsf)
getCol(PatternLibServer)
getCol(PsfDataServer)
getCol(TimeCorrector)

#undef getCol


void CalAtomCollectionServer::releaseColl(Xmm::Instrument instr,
										  unsigned n) const
{
	unsigned pos = index(instr, n);
	delete (*_atomsCollection)[pos];
	(*_atomsCollection)[pos] = 0;
} // CalAtomCollectionServer::releaseCollection


template<class AtomType>
void CalAtomCollectionServer::releaseCollection(const XmmInstrument *instr,
												AtomType *) const
{
	releaseColl(instr->id(), AtomType::id());
}

#define relCol(a)														\
template void CalAtomCollectionServer::releaseCollection(				\
												const XmmInstrument *,	\
												a *) const;

relCol(CanonicalCrossPsf)
relCol(AbsorptionDataServer)
relCol(BackgroundDataServer)
relCol(BadPixelDataServer)
relCol(Boresight)
relCol(CanonicalRedist)
relCol(CcdQuantumDataServer)
relCol(GeometryDataServer)
relCol(CtiCorrector)
relCol(DarkFrameDataServer)
relCol(EffectiveAreaDataServer)
relCol(EncircledEnergyDataServer)
relCol(EnergyCorrector)
relCol(FilterDataServer)
relCol(GratingDataServer)
relCol(HKvalRangesServer)
relCol(MiscDataServer)
relCol(ModeParameters)
relCol(OmColorTransformator)
relCol(OmGeoLinearizer)
relCol(OmFlatFieldServer)
relCol(OmFluxConverter)
relCol(OmPsf)
relCol(PatternLibServer)
relCol(PsfDataServer)
relCol(TimeCorrector)

#undef relCol

CalAtomCollectionServer::~CalAtomCollectionServer()
{
	for (unsigned n=0; n<_atomsCollection->size(); ++n)
		delete (*_atomsCollection)[n];
	delete _atomsCollection;
} // CalAtomCollectionServer::~CalAtomCollectionServer()


