//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CcdQuantumDataServer.cc --- implementation of CcdQuantumDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the CcdQuantumDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-17: 1.0.0: - first version
//		UL, 2000-05-20: 1.1.0: - many internal changes
//		UL, 2000-05-24: 1.2.0: - reading of PATTERNLIB is only done for EMOS1/2
//								 now (fixes SPR 1347)
//		UL, 2000-06-27: 1.3.0: - bug fix in CcdQuantumDataServer::at()
//		UL, 2000-06-30: 1.4.0: - small fix to bring in line with selectlib-4.35
//		UL, 2000-08-03: 1.5.0: - added RgsCcdQuantumDataServer
//		UL, 2000-08-29: 1.6.0: - if a pixel location is outside the FOV
//								 a constant QE=0 is returned; this fixes
//								 SPR: 1717/1736: Cannot get QE for chipy=1
//		UL, 2000-08-31: 1.7.0: - bug fix in RgsQuantumDataServer constructor
//								 (erroneous release of CAL atom)
//		UL, 2001-10-24: 1.8.0: - construction of RGS QE curves (Henke
//								 attenuation length data) amended by
//								 optional fine structure from RGS?_EXAFS CCF
//		UL, 2002-01-30: 1.9.0: - RgsCcdQuantumDataServer: added support for
//								 virtual H2O layer on RGS CCD to model
//								 observed absorption feature around O-edge
//								 in QE cuves
//		UL, 2002-11-20: 1.9.1: - fixed minor mistake in error message of
//								 EpicCcdQuantumDataServer::grade()
//
//	AUTHOR:
//		U. Lammers (UL), Thu Feb 17 14:13:14 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/CcdQuantumDataServer.cc,v 1.31 2002/11/20 15:40:19 ulammers Exp $
//

#include <cmath>
#include <algorithm>

#include "Tracer.h"
#include "CalCommon.h"
#include "CalOalUtility.h"
#include "stringutil.h"
#include "CalServer.h"
#include "Ccf.h"
#include "PatternLibServer.h"
#include "AbsorptionDataServer.h"
#include "PhysicalConstants.h"
#include "GeometryDataServer.h"
#include "CcdQuantumDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( CcdQuantumEfficiencyCurve )-------------------------------------------
//
CcdQuantumEfficiencyCurve &CcdQuantumEfficiencyCurve::operator+=(
										const CcdQuantumEfficiencyCurve &rhs)
{
	Slatec::CubicHermiteInterpolant<double>::operator+=(rhs);
	return *this;
} // CcdQuantumEfficiencyCurve::operator+=()


double CcdQuantumEfficiencyCurve::efficiency(double Elo, double Ehi) const
{
	if (fabs(Ehi-Elo) < 1.e-8)
		return this->efficiency(Elo);

	return this->integral(Elo, Ehi)/(Ehi-Elo);
} // CcdQuantumEfficiencyCurve::Efficiency


//
//	---( CcdQuantumDataServer )------------------------------------------------
//
CcdQuantumDataServer::CcdQuantumDataServer(Ccd ccd,
						 CalAtomCollection<CcdQuantumDataServer> *parent) :
		CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
		_ccd(ccd)
{
} // CcdQuantumDataServer::CcdQuantumDataServer()


const Proxy<PatternFractionDataServer>
CcdQuantumDataServer::patternFractionData() const
{
	errstr<<"Non-EPIC CcdQuantumDataServer cannot instantiate "
			"PatternFractionDataServer"<<Sas::error("UnspecializedObject");

	return Proxy<PatternFractionDataServer>(new PatternFractionDataServer());
} // CcdQuantumDataServer::patternFractionData()


CcdQuantumDataServer::~CcdQuantumDataServer()
{
} // CcdQuantumDataServer::~CcdQuantumDataServer()


//
//	---( QuantumDataRepository )-----------------------------------------------
//
template<class T>
class QuantumDataRepository {
public:
	QuantumDataRepository(const Table *table,
						  pair<unsigned short, unsigned short> rowRange,
						  const string &prefix,
						  const vector<string> &postfix,
						  const Proxy<vector<double> > &eGrid,
						  const T &outOfFOVdefaultItem)
		: _eGrid(eGrid), _regionIndex(-2),
		  _outOfFOVdefaultItem(outOfFOVdefaultItem)
	{
		ColumnData *cdr = table->column("REGION")->data();
		unsigned N = postfix.size();	//	number of grades/patterns
		//	loop over regions
		for (unsigned n=rowRange.first; n<=rowRange.second; ++n) {
			EvalTuple et(newEvalTuple(cdr, n));

			EfficiencyVector effVec(N);
			//	loop over grade/patterns
			for (unsigned i=0; i<N; ++i) {
				vector<double> qeVec;
				CalOalUtility::fillVector(table, prefix + '_' + postfix[i],
										  n, qeVec);
				effVec[i] = new T(*_eGrid, qeVec);
			}

			_localizedEfficiencyVectors.push_back(
										LocalizedEfficiencyVector(et, effVec));
		}
		delete cdr;		
	} // QuantumDataReposity()


	const QuantumDataRepository &at(unsigned short chipX,
									unsigned short chipY) const
	{
		_atCalled = true;

		//	search for first region applicable to rawX/rawY
		unsigned N = _localizedEfficiencyVectors.size();
		for (unsigned n=0; n<N; ++n) {
			const EvalTuple &et = _localizedEfficiencyVectors[n].first;
			if (et.chipX && et.chipY) {
				*et.chipX = DBL(chipX);
				*et.chipY = DBL(chipY);
			}
			if (et.evaluator->evalLogic()) {
				//	yup - found it
				_regionIndex = n;
				return *this;
			}
		}

		//
		//	could not find applicable region - check if the requested location
		//	is outside the FOV
		//
		int16 x = chipX,
			  y = chipY;
		CalInt16Vector chipXv(&x, 1),
					   chipYv(&y, 1);
		GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
		XmmPixelList pl = ct->chipCoordList(chipXv, chipYv);
		vector<bool> inFov = pl.insideFOV();
		if (inFov[0])
			//
			//	pixel is inside FOV but no applicable QE region could be found-
			//	this is a CCF problem
			//
			errstr<<"Could not find applicable region for rawX/rawY="
				  <<chipX<<'/'<<chipY<<" in quantum efficiency data retrieval"
				  <<Sas::error("NoQEdata");

		//	signal that pixel is outside FOV
		_regionIndex = -1;

		return *this;
	} // at()


	const T &item(unsigned short index) const
	{
		checkAt();
		assert(_regionIndex >= -1 &&
		   _regionIndex < INT(_localizedEfficiencyVectors.size()));

		if (_regionIndex >= 0) {
			const EfficiencyVector &effVec =
							_localizedEfficiencyVectors[_regionIndex].second;
			if (index >= effVec.size())
				errstr<<"Illegal pattern/grade ID "<<index<<" in selecting "
						"quantum efficiency/pattern fraction curve - "
						"valid range is [0, "<<effVec.size()-1<<"]"
					  <<Sas::error("IllegalPatternOrGrade");

			return *effVec[index];
		}

		//
		//	location is outside FOV - return d/f object
		//
		return _outOfFOVdefaultItem;
	} // item()


	~QuantumDataRepository()
	{
		for (unsigned n=0; n<_localizedEfficiencyVectors.size(); ++n) {
			delete _localizedEfficiencyVectors[n].first.evaluator;
			EfficiencyVector &effVec = _localizedEfficiencyVectors[n].second;
			unsigned N = effVec.size();
			for (unsigned nn=0; nn<N; delete effVec[nn++]);
			effVec.resize(0);
		}
	}

private:
	//	data members
	struct EvalTuple {
		const Evaluator	*evaluator;
		double			*chipX,
						*chipY;
	};
	typedef std::vector<T *>					EfficiencyVector;
	typedef pair<EvalTuple, EfficiencyVector>	LocalizedEfficiencyVector;
	std::vector<LocalizedEfficiencyVector>		_localizedEfficiencyVectors;

	Proxy<std::vector<double> >					_eGrid;
	mutable bool								_atCalled;
	mutable int									_regionIndex;
	T											_outOfFOVdefaultItem;

	//	methods
	void checkAt() const
	{
		if (!_atCalled) {
			unsigned chipX = calServer.instrument()->maxX()/2,
					 chipY = calServer.instrument()->maxY()/2;
			at(chipX, chipY);
			errstr<<"No chip location for QE specified - will retrieve data "
					"applicable at center ("<<chipX<<'/'<<chipY<<")"
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			_atCalled = false;
		}
	} // checkAt()


	EvalTuple newEvalTuple(const ColumnData *cdr, unsigned n) const
	{
		string region(stringColumnRow(cdr, n));

		if (!Evaluator::isValidExpression(region))
			errstr<<"The region specification `"<<region<<"' in `"
				  <<cdr->parent()->qualifiedName()<<"' is syntactically "
					"incorrect"<<Sas::error("CcfDataIntegrityError");

		EvalTuple et;
		string regStr = replaceall(lowercase(region),
									"polygon(", "polygon2(");
		et.evaluator = new Evaluator(region);
		NumericSymbols syms = et.evaluator->numericIdentifiers();
		et.chipX = et.chipY = 0;
		if (syms.name.size() == 2) {
			if (uppercase(syms.name[0]) == "RAWX" &&
				uppercase(syms.name[1]) == "RAWY") {
				et.chipX = syms.valuePtr[0];
				et.chipY = syms.valuePtr[1];
			} else if (uppercase(syms.name[0]) == "RAWY" &&
					   uppercase(syms.name[1]) == "RAWX") {
				et.chipX = syms.valuePtr[1];
				et.chipY = syms.valuePtr[0];
			} else
				errstr<<"Unrecognized symbols "<<syms.name[0]<<'/'
					  <<syms.name[1]<<" in `"
					  <<cdr->parent()->qualifiedName()
					  <<"' - expected RAWX + RAWY"
					  <<Sas::error("CcfDataIntegrityError");
		} else if (syms.name.size() != 0)
			errstr<<"Number of symbols in REGION specification `"
				  <<region<<"' from `"
				  <<cdr->parent()->qualifiedName()
				  <<"' must be two: RAWX + RAWY"
				  <<Sas::error("CcfDataIntegrityError");

		return et;
	} // newEvalTuple()
};


//
//	---( EpicCcdQuantumDataServer )-------------------------------------------
//
EpicCcdQuantumDataServer::EpicCcdQuantumDataServer(
										   Ccd ccd,
										   const Table *table,
										   const Proxy<vector<double> >&eGrid,
							 CalAtomCollection<CcdQuantumDataServer> *parent)
	: CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
	  CcdQuantumDataServer(ccd, parent)

{
	vector<unsigned> pattVec;
	for (unsigned col=0; col<table->columns(); ++col) {
		vector<string> sv = breakup(table->column(col)->name(), "_");
		if (sv.size() != 2 && sv[0] != "QE")
			continue;
		pattVec.push_back(toNumber<unsigned>(sv[1]));
	}
	sort(pattVec.begin(), pattVec.end());

	_patterns = pattVec.size();

	//
	//	construct _gradePatterns vector
	//	_gradePatterns[n] = list of patterns with grade n
	//
	if (calServer.instrument()->id() == Xmm::EMOS1||
		calServer.instrument()->id() == Xmm::EMOS2) {
		//	only EMOS comes with pattern lib
		PatternLibServer *pls = 0; pls = calServer.getAtom(pls, false);
		if (pls)
			_gradePatterns = pls->gradePatternList();
	}

	vector<string> pattIds(_patterns);
	for (unsigned n=0; n<_patterns; ++n)
		pattIds[n] = toString(n);

	_patternQECurves = new QuantumDataRepository<CcdQuantumEfficiencyCurve>(
							table, make_pair(0, table->rows()-1),
							"QE", pattIds, eGrid,
							CcdQuantumEfficiencyCurve(0.));
} // EpicCcdQuantumDataServer::EpicCcdQuantumDataServer()


const CcdQuantumEfficiencyCurve &EpicCcdQuantumDataServer::pattern(
													unsigned patternId) const
{
	return _patternQECurves->item(patternId);
} // EpicCcdQuantumDataServer::pattern()


CcdQuantumEfficiencyCurve EpicCcdQuantumDataServer::grade(unsigned gradeNr)
const
{
	if (gradeNr < 1 ||
		gradeNr > _gradePatterns.size())
		errstr<<"Invalid grade value "<<gradeNr<<" - must be in interval [1, "
			  <<_gradePatterns.size()<<"]"
			  <<Sas::error("IllegalGradeValue");

	const vector<unsigned short> &patternVec = _gradePatterns[gradeNr-1];
	if (!patternVec.size())
		errstr<<"There are no grade-"<<gradeNr<<" patterns"
			  <<Sas::error("NoSuchPattern");

	CcdQuantumEfficiencyCurve ec = _patternQECurves->item(patternVec[0]);
	for (unsigned n=1; n<patternVec.size(); ++n)
		ec += _patternQECurves->item(patternVec[n]);

	return ec;
} // EpicCcdQuantumEfficiencyCurve::grade()


const CcdQuantumDataServer &EpicCcdQuantumDataServer::at(
												   unsigned short chipX,
												   unsigned short chipY) const
{
	_patternQECurves->at(chipX, chipY);
	return *this;
} // EpicCcdQuantumDataServer::at()


const Proxy<PatternFractionDataServer>
EpicCcdQuantumDataServer::patternFractionData() const
{
	if (!_pattFracServer.operator->() || _pattFracServer->_mode !=
		calServer.state()->mode()->get())
		_pattFracServer =
			Proxy<PatternFractionDataServer>(new PatternFractionDataServer());

	return _pattFracServer;
} // EpicCcdQuantumDataServer::patternFractionData()


EpicCcdQuantumDataServer::~EpicCcdQuantumDataServer()
{
	delete _patternQECurves;
} // EpicCcdQuantumDataServer::~EpicCcdQuantumDataServer()


//
//	PatternFractionDataServer
//
namespace CombinedEventGrades {

bool contains(ID id, unsigned short grade)
{
	if (grade == 0 || grade > UNT(HIGHEST_PURE_GRADE)+1)
		return false;

	switch (id) {
		case SDTQ:	return grade <= QUADRUPLE+1; break;
		case SD	 :	return grade <= DOUBLE+1; break;
		default	 :  break;
	}

	return grade == id+1;
} // CombinedEventGrades::contains()

}

PatternFractionDataServer::PatternFractionDataServer() :
	_grade(CombinedEventGrades::INVALID)
{
	CifEntry *e = calServer.ccf()->entry(calServer.instrument(),
										 Ccf::QuantumEf);
	DataSet *dataSet = e->open();

	//	make sure we have a good CCF
	if (!dataSet->hasBlock("FRACTION_CHANNEL"))
		errstr<<"The present QuantumEf CCF `"<<dataSet->qualifiedName()
			  <<" does not contain pattern fraction data - please upgrade "
				" to a more recent version"
			  <<Sas::error("ImproperCcfConstituent");

	//	read energy axis
	ColumnData *cde =
					dataSet->table("EBINS_FRACTION")->column("ENERGY")->data();
	assert(cde->parent()->units() == "eV");
	real32 *energy = cde->real32Data();
	Proxy<vector<double> > eGrid =
						new vector<double>(energy, energy+cde->elements());
	delete cde;

	//	read total QE curve
	vector<double> qetotal;
	CalOalUtility::fillVector(dataSet->table("QE_TOTAL"), "QE_TOTAL", 0,
							  qetotal);
	_total = CcdQuantumEfficiencyCurve(*eGrid, qetotal);

	_mode = calServer.state()->mode()->get();
    unsigned short modeIndex = calServer.instrument()->modeIndex(_mode);

	//	read energy fraction data
	const Table *efracTab = dataSet->table("FRACTION_ENERGY");
	const char *gradePrefs[] = { "S", "D", "T", "Q", "SD", "SDTQ" };
	const vector<string> gradeIds(gradePrefs, gradePrefs+NUM_OF(gradePrefs));
	_fracQECurves = new QuantumDataRepository<CcdQuantumEfficiencyCurve>(
							efracTab, indexRange(efracTab, modeIndex),
							"FRAC", gradeIds, eGrid,
							CcdQuantumEfficiencyCurve(0.));

	//	read channel fraction data
	const Table *cfracTab = dataSet->table("FRACTION_CHANNEL");
	_fracQEvectors = new QuantumDataRepository<QEDataVector>(
							cfracTab, indexRange(cfracTab, modeIndex),
							"FRAC", gradeIds, eGrid,
							QEDataVector());
} // PatternFractionDataServer::PatternFractionDataServer()


pair<unsigned short, unsigned short> PatternFractionDataServer::indexRange(
											const Table *tab,
											unsigned short modeIndex) const
{
	vector<unsigned short> modes;
	typedef vector<unsigned short>::iterator Iter;
	CalOalUtility::fillVector(tab, "MODE_ID", modes);
	pair<Iter, Iter> pi = equal_range(modes.begin(), modes.end(), modeIndex);
	if (pi.first == modes.end() || *pi.first != modeIndex)
		errstr<<"QuantumEf CCF constituent `"<<tab->qualifiedName()
			  <<"' contains no data for instrument mode "
			  <<calServer.state()->mode()
			  <<Sas::error("NoSuchModeInCcf");

	return make_pair(pi.first-modes.begin(), pi.second-modes.begin());
} // PatternFractionDataServer::indexRange()


void PatternFractionDataServer::checkGrade() const
{
	if (_grade == CombinedEventGrades::INVALID)
		errstr<<"No grade selected - cannot retrieve pattern fraction data"
			  <<Sas::error("NoGradeSelected");
} // PatternFractionDataServer::chechGrade()


const CcdQuantumEfficiencyCurve &PatternFractionDataServer::total() const
{
	return _total;
} // PatternFractionDataServer::total()


const PatternFractionDataServer &PatternFractionDataServer::at(
													unsigned short chipX,
													unsigned short chipY) const
{
	_fracQECurves->at(chipX, chipY);
	_fracQEvectors->at(chipX, chipY);

	return *this;
} // PatternFractionDataServer::at()


const PatternFractionDataServer &PatternFractionDataServer::grade(
													EventGrades grades) const
{
	_grade = grades;

	return *this;
} // PatternFractionDataServer::grade()


const CcdQuantumEfficiencyCurve &PatternFractionDataServer::inEnergySpace()
const
{
	checkGrade();

	return _fracQECurves->item(_grade);
} // PatternFractionDataServer::inEnergySpace()


const vector<double> &PatternFractionDataServer::inChannelSpace() const
{
	checkGrade();

	return _fracQEvectors->item(_grade);
} // PatternFractionDataServer::inChannelSpace()


PatternFractionDataServer::~PatternFractionDataServer()
{
	delete _fracQECurves;
	delete _fracQEvectors;
} // PatternFractionDataServer::~PatternFractionDataServer()


//
//	RGS
//
RgsCcdQuantumDataServer::RgsCcdQuantumDataServer(
					     Ccd ccd,
						 const Table *materialThicknessTable,
						 const Table *SiThicknessTable,
						 CalAtomCollection<CcdQuantumDataServer> *parent)
	: CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
	  CcdQuantumDataServer(ccd, parent), _currentSiThickness(0.)
{
	//
	//	get thicknesses of material layers
	//
	ColumnData *cdccdid = materialThicknessTable->column("CCDID")->data(),
			   *cdal = materialThicknessTable->column("D_AL")->data(),
			   *cdmgf2 = materialThicknessTable->column("D_MGF2")->data(),
			   *cdsi = materialThicknessTable->column("D_SI")->data(),
			   *cdsio2 = materialThicknessTable->column("D_SIO2")->data(),
			   *cdalo2 = materialThicknessTable->column("D_ALO2")->data(),
			   *cdh2o = 0;
	if (materialThicknessTable->hasColumn("D_H2O"))
		cdh2o = materialThicknessTable->column("D_H2O")->data();

	int8 *ccdid = cdccdid->int8Data();
	real32 *al = cdal->real32Data(),
		   *mgf2 = cdmgf2->real32Data(),
		   *si = cdsi->real32Data(),
		   *sio2 = cdsio2->real32Data(),
		   *alo2 = cdalo2->real32Data(),
		   *h2o = cdh2o ? cdh2o->real32Data() : 0; // old CCF are lacking this
												   // column

	unsigned N = materialThicknessTable->rows();
	unsigned n;
	for (n=0; n<N; ++n)
		if (ccdid[n] == ccd)
			break;
	if (n>=N)
		errstr<<"No entry for CCD "<<ccd<<" in `"
			  <<materialThicknessTable->qualifiedName()
			  <<Sas::error("CcfIntegrityError");

	double dAl = 1.e-7 * al[n],		//	nm -> cm
		   dMgF2 = 1.e-7 * mgf2[n],
		   dSi = 1.e-7 * si[n],
		   dSiO2 = 1.e-7 * sio2[n],
		   dAl2O3 = 1.e-7 * alo2[n],
		   dH2O = 1.e-7 * (h2o ? h2o[n] : 0.);

	delete cdccdid;
	delete cdal;
	delete cdmgf2;
	delete cdsi;
	delete cdsio2;
	delete cdalo2;
	delete cdh2o;

	//
	//	construct Si thickness map
	//
	double dActiveSi = SiThicknessTable->attribute("MTHICK")->asReal();
	if (SiThicknessTable->hasColumn("CHIPX")) {
		ColumnData *cdchipX = SiThicknessTable->column("CHIPX")->data(),
				   *cdchipY = SiThicknessTable->column("CHIPY")->data(),
				   *cdfringes = SiThicknessTable->column("FRINGES")->data();
		_thicknessSi = new ThicknessSiInterpolator(SiThicknessTable->rows(),
												   cdchipX->int16Data(),
												   cdchipY->int16Data(),
												   cdfringes->int16Data(),
												   dActiveSi);
		delete cdchipX;
		delete cdchipY;
		delete cdfringes;
	} else
		_thicknessSi = new ThicknessSiInterpolator(dActiveSi);

	//
	//	construct transmission / tau*henke curves
	//
	AbsorptionDataServer *ads = 0; ads = calServer.getAtom(ads);
    const AbsCoefCurve &f2Al    = ads->henke("Al").f2(),
                       &f2F     = ads->henke("F").f2(),
                       &f2Mg    = ads->henke("Mg").f2(),
                       &f2O     = ads->henke("O").f2(),
                       &f2Si    = ads->henke("Si").f2();
	double	AAl		= ads->atomicData("Al").weight(),			// [g/mol]
			AF		= ads->atomicData("F").weight(),
			AMg		= ads->atomicData("Mg").weight(),
			AO		= ads->atomicData("O").weight(),
			ASi		= ads->atomicData("Si").weight(),
			AH		= 1.,

			rhoAl	= ads->atomicData("Al").density() * 1.e-3,	// [g/cm^3]
			rhoSi	= ads->atomicData("Si").density() * 1.e-3;

	double	AMgF2	= AMg + 2. * AF,
			rhoMgF2	= 3.15,	//	[g/cm^3]
			AAl2O3	= 2. * AAl + 3. * AO,
			rhoAl2O3= 3.97,	//	[g/cm^3]
			ASiO2	= ASi + 2. * AO,
			AH2O	= 2. * AH + AO,
			rhoSiO2	= 2.64,	//	[g/cm^3]
			rhoH2O	= 1.;	//	[g/cm^3]

	char *absMaterial[] = { "MgF2",	"Al", "Al2O3", "SiO2", "Si", "H2O", "Si" };
	bool  passive[]		= { true,   true, true,    true,   true, true, false };
	double d[]			= { dMgF2,  dAl,  dAl2O3,  dSiO2,  dSi,  dH2O, 0. };

	_eGrid = f2Si.abscissa();
	unsigned Ne = _eGrid.size(),
			 Nmat = NUM_OF(absMaterial);
	vector<double> muInv[Nmat];
	for (unsigned n=0; n<Nmat; muInv[n++].resize(Ne));
	for (unsigned n=0; n<Ne; ++n) {
		double &eV	= _eGrid[n],								// [eV]
				f0	= 2. * PhysicalConstants::avogadro *		// [1/mol]
					  PhysicalConstants::r0 * 1.e2 *			// [cm]
					  PhysicalConstants::eVtimesAo/eV * 1.e-8;	// [cm]
		muInv[0][n]	= f0*rhoMgF2/AMgF2*(f2Mg(eV)+2.*f2F(eV));	// MgF2
		muInv[1][n]	= f0*rhoAl/AAl*f2Al(eV);					// Al
		muInv[2][n]	= f0*rhoAl2O3/AAl2O3*3.*f2O(eV);			// Al2O3
		muInv[3][n]	= f0*rhoSiO2/ASiO2*(f2Si(eV)+2.*f2O(eV));	// SiO2
		muInv[4][n]	= f0*rhoSi/ASi*f2Si(eV);					// Si
		muInv[5][n] = f0*rhoH2O/AH2O*(2.*0.+f2O(eV));			// H20; neglect
																// f2(H)
		muInv[6][n]	= muInv[4][n];								// Si (active)
	}

	//	construct interpolated curves
	InterpolatedCurve invAbs[Nmat];
	for (unsigned n=0; n<Nmat; ++n)
		invAbs[n] = InterpolatedCurve(_eGrid, muInv[n]);

	//	augment inverse absorption lengths by EXAFS data from CCF
	RgsEXAFSDataServer exafs(ccd);
	for (unsigned n=0; n<Nmat; ++n)
		if (exafs.hasMaterial(absMaterial[n], passive[n])) {
			const vector<vector<double> > &E =
							exafs.material(absMaterial[n], passive[n]).E(),
									&mu =
							exafs.material(absMaterial[n], passive[n]).invMu();
			unsigned Ne = E.size();
			for (unsigned i=0; i<Ne; ++i) {
				//	update absorption lengths in material layer <n> by
				//	EXAFS data ...
				invAbs[n].replace(E[i], mu[i]);

				SAS_DEBUG(cout<<"Absorption lengths for layer "<<n
							  <<" thickness: "<<d[n]<<endl;
						  for (unsigned nn=0; nn<E[i].size(); ++nn)
							cout<<E[i][nn]<<' '<<1./mu[i][nn]<<endl;)

				//	and add new energy points to other E-axes
				for (unsigned j=0; j<Nmat; ++j) {
					if (j==n)
						continue;
					invAbs[j].add(E[i]);
				}
			}
		}

	_eGrid = invAbs[0].abscissa();
	_tau = invAbs[Nmat-1];

	//	now construct transmission probabilities through inactive layer
	Ne = _eGrid.size();
	vector<double> trans(Ne);
	SAS_DEBUG(cout<<"Transmission probabilities:"<<endl;)
	for (unsigned n=0; n<Ne; ++n) {
		trans[n] = 0.;
		for (unsigned i=0; i<Nmat-1; ++i)
			trans[n] += d[i]*invAbs[i](_eGrid[n]);
		trans[n] = exp(-trans[n]);
		SAS_DEBUG(cout<<_eGrid[n]<<' '<<trans[n]<<endl;)
	}

	_transmission = InterpolatedCurve(_eGrid, trans);

	//
	//	QE(E) = (1-exp(-_tau(E)*dSi)*_transmission(E))
	//
} // RgsCcdQuantumDataServer::RgsCcdQuantumDataServer()


const CcdQuantumEfficiencyCurve &RgsCcdQuantumDataServer::pattern(
														unsigned patternId)
const
{
	static CcdQuantumEfficiencyCurve _grade1Curve;
	if (patternId)
		errstr<<"Can only give pattern-0 QE curves"
			  <<Sas::error("NoSuchQEcurve");
	_grade1Curve = grade(1);

	return _grade1Curve;
} // RgsCcdQuantumDataServer::pattern()


CcdQuantumEfficiencyCurve RgsCcdQuantumDataServer::grade(unsigned gradeNr)
const
{
	if (gradeNr != 1)
		errstr<<"QE only available for grade-1 events"
			  <<Sas::error("NoSuchQE");

	if (_currentSiThickness == 0.) {
		errstr<<"Computing average Si-layer thickness for QE calculation"
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		averageSiThickness();
	}

	unsigned N = _eGrid.size();
	vector<double> v(N);
	for (unsigned n=0; n<N; ++n) {
		double e = _eGrid[n];
		v[n] = (1. - exp(-_currentSiThickness * _tau(e))) * _transmission(e);
	}

	return CcdQuantumEfficiencyCurve(_eGrid, v);
} // RgsCcdQuantumEfficiencyCurve::grade()


void RgsCcdQuantumDataServer::averageSiThickness() const
{
	unsigned short minX = calServer.instrument()->minX(),
				   maxX = calServer.instrument()->maxX(),
				   minY = calServer.instrument()->minY(),
				   maxY = calServer.instrument()->maxY();

	unsigned n=0;
	for (unsigned ix=minX; ix<=maxX; ix+=10)
		for (unsigned iy=minY; iy<=maxY; iy+=10) {
			++n;
			_currentSiThickness += (*_thicknessSi)(ix, iy);
		}

	_currentSiThickness /= DBL(n);
} // RgsCcdQuantumDataServer::averageSiThickness()


RgsCcdQuantumDataServer::RgsEXAFSDataServer::RgsEXAFSDataServer(Ccd ccd)
	: _materialSelected(-1)
{
	if (CifEntry *cifEntry = calServer.ccf()->entry(calServer.instrument(),
												Ccf::EXAFS, false)) {
		DataSet *set = cifEntry->open();
		//	loop over all blocks
		unsigned nBlocks = set->blocks();
		for (unsigned n=0; n<nBlocks; ++n) {
			Table *table = set->table(n);
			vector<string> v(breakup(table->name(), "-"));
			if (v.size() != 3 ||
				v[1].length() != 1 || (v[1][0] != 'P' && v[1][0] != 'A') ||
				v[2].length() != 4)
				errstr<<"Unexpected CCF table name `"<<table->name()
					  <<"'"<<Sas::error("UnexpectedCcfTableName");
			if (toNumber<unsigned>(v[2].substr(3, 1)) != ccd)
				continue;

			unsigned N = table->rows();
			assert(table->column("ENERGY")->units() == "eV");
			assert(table->column("MABSLEN")->units() == "cm");
			vector<vector<double> > e, f;
			for (unsigned n=0; n<N; ++n) {
				vector<double> e0, f0;
				CalOalUtility::fillVector(table, "ENERGY", n, e0);
				CalOalUtility::fillVector(table, "MABSLEN", n, f0);
				transform(f0.begin(), f0.end(), f0.begin(),
						  CalFunctors::Inverse<double>());
													//	need inverse absorption
													//	lengths
				e.push_back(e0);
				f.push_back(f0);
			}
			_knownMaterials.push_back(make_pair<string, bool>(v[0],
															  v[1][0] == 'P'));
			_eGrids.push_back(e);
			_invMu.push_back(f);
		}
		dataSetServer->close(set);
	}
} // RgsEXAFSDataServer::RgsEXAFSDataServer()


bool RgsCcdQuantumDataServer::RgsEXAFSDataServer::hasMaterial(
												const string &mat,
												bool isPassiveLayer) const
{
	return find(_knownMaterials.begin(), _knownMaterials.end(),
				make_pair<string, bool>(mat, isPassiveLayer)) != 
														_knownMaterials.end();
} // RgsEXAFSDataServer::hasMaterial()


const RgsCcdQuantumDataServer::RgsEXAFSDataServer &
RgsCcdQuantumDataServer::RgsEXAFSDataServer::material(
												const string &mat,
												bool isPassiveLayer) const
{
	vector<pair<string, bool> >::const_iterator p =
			find(_knownMaterials.begin(), _knownMaterials.end(),
				make_pair<string, bool>(mat, isPassiveLayer));
	if (p == _knownMaterials.end())
		errstr<<"No EXAFS data for "<<(isPassiveLayer ? "passive" : "active")
			  <<" material `"<<mat<<"' read from CCF"
			  <<Sas::error("NoEXAFSdata");

	unsigned pos = p - _knownMaterials.begin();
	_materialSelected = pos;

	return *this;
} // RgsEXAFSDataServer::material()


void RgsCcdQuantumDataServer::RgsEXAFSDataServer::checkMaterialSelected() const
{
	if (_materialSelected < 0)
		errstr<<"Select first material with "
				"RgsEXAFSDataServer::material()"
			  <<Sas::error("NoMaterialSelected");

	assert(_materialSelected >= 0 && _materialSelected < INT(_eGrids.size()));
} // RgsEXAFSDataServer::materialSelected()


const vector<vector<double> > &RgsCcdQuantumDataServer::RgsEXAFSDataServer::E()const
{
	checkMaterialSelected();
	return _eGrids[_materialSelected];
} // RgsEXAFSDataServer::E()


const vector<vector<double> >&
	RgsCcdQuantumDataServer::RgsEXAFSDataServer::invMu() const
{
	checkMaterialSelected();
	return _invMu[_materialSelected];
} // RgsEXAFSDataServer::invMu()


const CcdQuantumDataServer &RgsCcdQuantumDataServer::at(unsigned short chipX,
														unsigned short chipY)
const
{
	_currentSiThickness = (*_thicknessSi)(chipX, chipY);

	return *this;
} // RgsCcdQuantumDataServer::at()


RgsCcdQuantumDataServer::~RgsCcdQuantumDataServer()
{
	delete _thicknessSi;
} // RgsCcdQuantumDataServer::~RgsCcdQuantumDataServer()



CalAtomCollection<CcdQuantumDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	CcdQuantumDataServer *null = 0;
	_collection = new vector<CcdQuantumDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::QuantumEf);
	DataSet *dataSet = e->open();

	switch(instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
		case Xmm::EPN: {
			//	read energy grid
			CcfConstituentTable ccftab(dataSet, "EBINS");
			const Table *table = ccftab.block();
			ColumnData *cde = table->column("ENERGY")->data();
			real32 *energy = cde->real32Data();
			Proxy<vector<double> > eGrid = new vector<double>(energy,
													  energy+cde->elements());
			delete cde;

			for (unsigned n=0; n<_collection->size(); ++n) {
				CcdQuantumDataServer *&qeds = (*_collection)[n];

				string tableName = "QE_CCD" + toString(n+1);
				if (!dataSet->hasBlock(tableName)) {
					errstr<<"No QE data for CCD #"<<n+1<<" in "
						  <<dataSet->name()<<Sas::warning("NoQEdata");
					continue;
				}

				CcfConstituentTable ccftab(dataSet, tableName);
				const Table *table = ccftab.block();
				qeds = new EpicCcdQuantumDataServer(n+1, table, eGrid, this);
			}
			break; }
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CcfConstituentTable ccftab(dataSet, "CCD_DESC");
			const Table *materialThicknessTable = ccftab.block();
			for (unsigned n=0; n<_collection->size(); ++n) {
				CcdQuantumDataServer *&qeds = (*_collection)[n];
				string tableName = "SI" + toString(n+1);
				if (!dataSet->hasBlock(tableName)) {
					errstr<<"No QE data for CCD #"<<n+1<<" in "
						  <<dataSet->name()<<Sas::warning("NoQEdata");
					continue;
				}

				CcfConstituentTable ccftab(dataSet, tableName);
				const Table *table = ccftab.block();
				qeds = new RgsCcdQuantumDataServer(n+1, materialThicknessTable,
												   table, this);
			}			
			break; }
		default:
			break;
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<CcdQuantumDataServer>::CalAtomCollection()



CcdQuantumDataServer *CalAtomCollection<CcdQuantumDataServer>::select(
					Ccd ccd,
					Node,
					unsigned) const
{
	return (*_collection)[ccd-1];
} // CalAtomCollection<CcdQuantumDataServer>::select()
