//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EffectiveAreaDataServer.cc --- implementation of EffectiveAreaDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the EffectiveAreaDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//		UL, 2000-04-25: 1.1.0: - temporarily disabled instantiation of
//								 low-accuracy EffectiveAreaDataServer
//								 (numerical problems with 2-D polynomial fit
//								 from IDL)
//		UL, 2000-05-09: 1.2.0: - bug fix in
//									EffectiveAreaDataServer::vignettingCurve
//								 for RGS
//		UL, 2000-09-15: 1.2.1: - theta>0 was implicitely assumed in
//									effectiveAreaCurve()
//		UL, 2001-11-01: 1.3.0: - empirical Ae curve correction for RGS;
//								 driven by number in
//								 RGS?_QUANTUMEF:RGA_EFFAREACORR
//		UL, 2002-02-12: 1.4.0: - MediumAccuracyEffectiveAreaDataServer::
//									effectiveArea(): interpolate in energy
//								 with fourth-order polynomial
//		UL, 2002-03-14: 1.5.0: - modify Ae curve by interpolated empirical
//								 correction data (instead of supplementing
//								 Ae data vector)
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 18:24:59 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/EffectiveAreaDataServer.cc,v 1.14 2002/08/19 12:19:24 ulammers Exp $
//

#include <cmath>

#include "Tracer.h"
#include "CalOalUtility.h"
#include "CalServer.h"
#include "EffectiveAreaDataServer.h"
#include "GratingDataServer.h"
#include "Interpolator.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( EffectiveAreaCurve )-----------------------------------------------
//
double EffectiveAreaCurve::area(double E) const
{
	return (*this)(E);
} // EffectiveAreaCurve::area()


double EffectiveAreaCurve::area(double Elo, double Ehi) const
{
	if (fabs(Ehi-Elo) < 1.e-8)
		return this->area(Elo);

	return this->integral(Elo, Ehi)/(Ehi-Elo);
} // EffectiveAreaCurve::area()


//
//	---( EffectiveAreaDataServer )-------------------------------------------
//
EffectiveAreaDataServer::EffectiveAreaDataServer(
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent)
{
} // EffectiveAreaDataServer::EffectiveAreaDataServer()


void EffectiveAreaDataServer::dump(ostream &)
{
	errstr<<"Sorry dumping EffectiveAreaDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");
} // EffectiveAreaDataServer::dump()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurveFull(
															double theta,
															double phi) const
{
	return effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(double theta,
														double phi) const
{
	errstr<<"Generic EffectiveAreaDataServer cannot provide effective area "
			"curve"<<Sas::error("MeaningLessRequest");

	return effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


double EffectiveAreaDataServer::effectiveArea(double E,
											  double theta, double phi) const
{
	return effectiveAreaCurve(theta, phi)(E);
} // EffectiveAreaDataServer::effectiveArea()



EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(int,
														double theta,
														double phi) const
{
	errstr<<"Generic EffectiveAreaDataServer cannot provide RGS-specific "
			"effective area curve"<<Sas::error("MeaninglessRequest");

	return effectiveAreaCurve(theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve EffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
														int order,
														double theta,
														double phi) const
{
	return effectiveAreaCurve(order, theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()


VignettingCurve EffectiveAreaDataServer::vignettingCurve(double theta,
												  double phi) const
{
	VignettingCurve vc(effectiveAreaCurve(theta, phi));
	vc /= effectiveAreaCurve(0., 0.);

	return vc;
} // EffectiveAreaDataServer::vignettingCurve()


VignettingCurve EffectiveAreaDataServer::vignettingCurve(int spectralOrder,
												  double theta,
												  double phi) const
{
	VignettingCurve vc(effectiveAreaCurve(spectralOrder, theta, phi));
	vc /= effectiveAreaCurve(spectralOrder, 0., 0.);

	return vc;
} // EffectiveAreaDataServer::vignettingCurve()


EffectiveAreaDataServer::~EffectiveAreaDataServer()
{
} // EffectiveAreaDataServer::~EffectiveAreaDataServer()


//
//	---( LowAccuracyEffectiveAreaDataServer )----------------------------------
//
LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer(
								Proxy<EffectiveAreaCurve>
									onAxisEffectiveAreaCurve,
								const vector<double> &eGrid,
								const vector<DImage> &coefs,
								double deltaF,
								double deltaY,
								double deltaZ,
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
	  EffectiveAreaDataServer(parent),
	  _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve),
	  _eGrid(eGrid),
	  _deltaF(deltaF), _deltaY(deltaY), _deltaZ(deltaZ)
{
	MiscDataServer *mds = 0; mds = calServer.getAtom(mds);
	_focalLength = mds->get("FOCAL_LENGTH");

	//	compute coefficient image
	unsigned N = _eGrid.size();
	if (N != coefs.size())
		errstr<<"Size mismatch between energy grid and coefficient image "
				"vector"
			  <<Sas::error("SizeMismatch");
	if (N < 2)
		errstr<<"Need at least two elements in energy grid vector"
			  <<Sas::error("TooLittleDataPoints");

	unsigned jMax = coefs[0].xExtent(),
			 kMax = coefs[0].yExtent();

	_coefsCurve = Image<Slatec::CubicHermiteInterpolant<double> *>(jMax, kMax);
	for (unsigned j=0; j<jMax; ++j)
		for (unsigned k=0; k<kMax; ++k) {
			vector<double> v(N);
			for (unsigned n=0; n<N; ++n)
				v[n] = coefs[n][j][k];
			_coefsCurve[j][k] =
					new Slatec::CubicHermiteInterpolant<double>(_eGrid, v);
		}
} // LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer()


inline pair<double, double> LowAccuracyEffectiveAreaDataServer::yz(
													double theta, double phi)
const
{
	double	c = _focalLength*(1+_deltaF/_focalLength),
			tanTheta = tan(theta/3600. * DEG_TO_RAD);

	return make_pair(c * tanTheta * cos(phi) + _deltaY,
					 c * tanTheta * sin(phi) + _deltaZ);
} // LowAccuracyEffectiveAreaDataServer::yz()


inline double LowAccuracyEffectiveAreaDataServer::vignettingFactor(
									double E, const pair<double, double> &yz,
								    unsigned jMax, unsigned kMax) const
{
	double ypow = 1.,
		   zpow = 1.,
		   v = 0.;

	for (unsigned j=0; j<jMax; ++j) {
		for (unsigned k=0; k<kMax; ++k) {
			v += (*_coefsCurve[j][k])(E) * ypow * zpow;
			ypow *= yz.first;
		}
		zpow *= yz.second;
	}

	return v;
} // LowAccuracyEffectiveAreaDataServer::vignettingFactor()


EffectiveAreaCurve LowAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
										double theta,	//	[arcsec]
										double phi)		//	[rad]
const
{
	pair<double, double> yz = this->yz(theta, phi);
	EffectiveAreaCurve ae(*_onAxisEffectiveAreaCurve);
	const vector<double> &E = ae.abscissa();
	unsigned N = E.size();
	vector<double> v(N);
	unsigned jMax = _coefsCurve.xExtent(),
			 kMax = _coefsCurve.yExtent();

	for (unsigned n=0; n<N; ++n)
		v[n] = vignettingFactor(E[n], yz, jMax, kMax);

	return ae*Slatec::CubicHermiteInterpolant<double>(E, v);
} // LowAccuracyDataServer::effectiveAreaCurve()


double LowAccuracyEffectiveAreaDataServer::effectiveArea(double E,
														 double theta,
														 double phi) const
{
	unsigned jMax = _coefsCurve.xExtent(),
			 kMax = _coefsCurve.yExtent();
	return (*_onAxisEffectiveAreaCurve)(E) *
							vignettingFactor(E, yz(theta, phi), jMax, kMax);
} // LowAccuracyEffectiveAreaDataServer::effectiveArea()


LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()
{
	unsigned jMax = _coefsCurve.xExtent(),
			 kMax = _coefsCurve.yExtent();
	for (unsigned j=0; j<jMax; ++j)
		for (unsigned k=0; k<kMax; ++k)
			delete _coefsCurve[j][k];
} // LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()



MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataServer(
							Proxy<EffectiveAreaCurve>
								onAxisEffectiveAreaCurve,
							const Table *vignTable,
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
	  EffectiveAreaDataServer(parent),
	  _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve)
{
	const ColumnData *cdE = vignTable->column("ENERGY")->data(),
				 *cdVignFac = vignTable->column("VIGNETTING_FACTOR")->data();
	real32 *E = cdE->real32Data(),
		   *vignFac = cdVignFac->real32Data();

	unsigned Ne = vignTable->rows(),
			 Ntheta = cdVignFac->elements();
	_eGrid.resize(Ne);
	vector<double> thetaGrid(Ntheta);
	double deltaTheta = vignTable->attribute("D_THETA")->asReal() // [deg]
						* 3600.; // [deg->arcsec]
	for (unsigned n=0; n<Ntheta; ++n)
		thetaGrid[n] = DBL(n) * deltaTheta;
	thetaGrid.push_back(1.e4);	//	one data point at extreme angle to avoid
								//	extrapolation artefacts

	vector<double> index(Ne);
	for (unsigned ne=0; ne<Ne; ++ne) {
		index[ne] = ne;
		_eGrid[ne] = E[ne];
		vector<double> fac(Ntheta);
		for (unsigned nt=0; nt<Ntheta; ++nt)
			fac[nt] = vignFac[cdVignFac->position(ne, nt)];
		fac.push_back(0.);	//	vignetting 0 at theta=1.e4"

		_vignetting.push_back(new Slatec::CubicHermiteInterpolant<double>
									(thetaGrid, fac));
	}
	_vignettingIndexLocator = new Slatec::CubicHermiteInterpolant<double>
									(_eGrid, index);

	delete cdE;
	delete cdVignFac;

#if 0
	unsigned N = eGrid.size();
	if (flatFields.size() != N)
		errstr<<"Vector with flatfields and energy grid vector have different "
				"sizes"<<Sas::error("SizeMismatch");

	//
	//	construct the interpolation maps
	//
	for (unsigned n=0; n<N; ++n) {
		const SImage &map = flatFields[n];
		int xExt = map.xExtent(),
			yExt = map.yExtent();
		vector<float> x(xExt),
					  y(yExt),
					  z(map.data(), map.data()+ xExt * yExt);
		for (int i=0; i<xExt; ++i)
			x[i] = DEG_TO_RAD*((i-INT(refPixXY)+1)*deltaXY + refValXY);
		for (int j=0; j<yExt; ++j)
			y[j] = DEG_TO_RAD*((j-INT(refPixXY)+1)*deltaXY + refValXY);

		_flatFields.push_back(
			new Slatec::Bivariate2DInterpolantRectangular<float>(x, y, z));
	}
#endif
} // MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataSercer()


EffectiveAreaCurve MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
											double theta,
											double) const
{
	unsigned N = _eGrid.size();
	vector<double> vignetting(N);
	double abstheta = fabs(theta);
	for (unsigned n=0; n<N; ++n)
		vignetting[n] = (*_vignetting[n])(abstheta);

	Slatec::CubicHermiteInterpolant<double> vignFunc(_eGrid, vignetting),
										onAxisAe(*_onAxisEffectiveAreaCurve);

	return onAxisAe * vignFunc;
} // MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve()


inline unsigned MediumAccuracyEffectiveAreaDataServer::vignettingIndex(
																double E) const
{
	int i = CalOalUtility::closestInt((*_vignettingIndexLocator)(E)),
		imax = _eGrid.size()-1;
	--i;

	if (i<0)
		return 0;
	if (i+3>imax)
		return UNT(imax-3);

	return UNT(i);
} // MediumAccuracyEffectiveAreaDataServer::vignettingIndex()


double MediumAccuracyEffectiveAreaDataServer::effectiveArea(
									double E, double theta, double) const
{
	unsigned i = vignettingIndex(E);
	double ft = fabs(theta);
	double xp[] = { _eGrid[i], _eGrid[i+1], _eGrid[i+2], _eGrid[i+3] },
		   yp[][1] = { {(*_vignetting[i])(ft)}, {(*_vignetting[i+1])(ft)},
					{(*_vignetting[i+2])(ft)}, {(*_vignetting[i+3])(ft)} },
		   vig[0];

	Interpolator<1>::doPoly4(xp, yp, E, vig);

	SAS_DEBUG(cout<<xp[0]<<' '<<yp[0][0]<<endl;
			  cout<<xp[1]<<' '<<yp[1][0]<<endl;
			  cout<<xp[2]<<' '<<yp[2][0]<<endl;
			  cout<<xp[3]<<' '<<yp[3][0]<<endl;
			  cout<<"vignetting factor: "<<vig[0]<<endl;)

	return (*_onAxisEffectiveAreaCurve)(E) * vig[0];
} // MediumAccuracyEffectiveAreaDataServer::effectiveArea()


MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()
{
	unsigned N = _vignetting.size();
	for (unsigned n=0; n<N; ++n)
		delete _vignetting[n];
	delete _vignettingIndexLocator;
} // MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()


//
//	-----( EMOS/RGS EffectiveAreaDataServer )---------------------------------
//
EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer(
						Proxy<EffectiveAreaDataServer> mds,
						CalAtomCollection<EffectiveAreaDataServer> *parent)
		: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
		  EffectiveAreaDataServer(parent),
		  _unvignetted(mds)
{
	GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
	_gds = gds;
} // EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurveFull(
																double theta,
																double phi)
const
{
	return _unvignetted->effectiveAreaCurve(theta, phi);
} // EmosEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurve(
															double theta,
															double phi) const
{
	InterpolatedCurve oef = _gds->obscurationEfficiency(theta, phi);
	EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));
	ae *= (1. - oef);

	return ae;
} // EmosEffectiveAreaDataServer::effectiveAreaCurve()


double EmosEffectiveAreaDataServer::effectiveArea(double E,
												  double theta, double phi)
const
{
	return _unvignetted->effectiveArea(E, theta, phi) *
							(1. - _gds->obscurationEfficiency(E, theta, phi));
} // EmosEffectiveAreaDataServer::effectiveArea()


RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer(
						Proxy<EffectiveAreaDataServer> mds,
						CalAtomCollection<EffectiveAreaDataServer> *parent)
		: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
		  EffectiveAreaDataServer(parent),
		  _unvignetted(mds)
{
	//
	//	empirical Ae correction from table RGS?_QUANTUMEF.CCF:RGA_EFFAREACORR
	//
	CifEntry *qe = calServer.ccf()->entry(calServer.instrument(),
										  Ccf::QuantumEf);
	DataSet *set = qe->open();
	if (set->hasBlock("RGA_EFFAREACORR")) {
		//	table is present - read contents
		Table *table = set->table("RGA_EFFAREACORR");
		std::vector<double> e, f;
		CalOalUtility::fillVector(table, "ENERGY", _empiricalCorrectionEgrid);
		CalOalUtility::fillVector(table, "FACTOR", _empiricalCorrectionValue);
	}
	dataSetServer->close(set);
} // RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurveFull(
																double theta,
																double phi)
const
{
	return _unvignetted->effectiveAreaCurve(theta, phi);
} // RgsEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurve(
															int spectralOrder,
															double theta,
															double phi) const
{
	EffectiveAreaCurve ae(intrinsicEffectiveAreaCurve(spectralOrder,
													  theta, phi));

	GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
	ae *= gds->selfVignettingEfficiency(ae.abscissa(), theta, phi,
										spectralOrder);

	//
	//	[2]: missing are last two factors:
	//		CAL_rgsCrossPsf x f_xdspsel
	//		CAL_rgsLSF x f_prdersel
	//

	return ae;
} // RgsEffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
														   int spectralOrder,
														   double theta,
														   double phi) const
{
	EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));

	GratingDataServer *gds = 0; gds = calServer.getAtom(gds);
	ae *= gds->interceptEfficiency(theta, phi);
	ae *= gds->reflectionEfficiency(theta, phi, spectralOrder);

	//	apply overall empirical correction curve
	unsigned N = _empiricalCorrectionEgrid.size();
	if (N) {
		//	modify Ae curve by interpolated empirical correction data
		return ae *
			Slatec::CubicHermiteInterpolant<double>(_empiricalCorrectionEgrid,
													_empiricalCorrectionValue);
//
//	old scheme
//		vector<double> vadd(N);
//		for (unsigned n=0; n<N; ++n)
//			vadd[n] = ae(_empiricalCorrectionEgrid[n]) *
//												_empiricalCorrectionValue[n];
//		ae.supplement(_empiricalCorrectionEgrid, vadd);
	}

	return ae;
} // RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve()



CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	EffectiveAreaDataServer *null = 0;
	_collection = new vector<EffectiveAreaDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::XAreaEf);
	DataSet *set = e->open();

	//
	//	read on-axis effective area data
	//
	CcfConstituentTable ccftab(set, "ONAXISXAREAEF");
	const Table *table = ccftab.block();
	unsigned N = table->rows();
	vector<double> Evec(N),
				   onAxisAe(N);
	ColumnData *cdE = table->column("ENERGY")->data(),
			   *cdAe = table->column("AREA")->data();
	real32 *E = cdE->real32Data(),
		   *area = cdAe->real32Data();
	for (unsigned n=0; n<N; ++n) {
		Evec[n] = E[n];
		onAxisAe[n] = area[n];
	}
	delete cdE;
	delete cdAe;

	EffectiveAreaDataServer *mds;
	Proxy<EffectiveAreaCurve> ae = new EffectiveAreaCurve(Evec, onAxisAe);

	AccuracyLevel acc = MEDIUM;	//	for the moment disable low-accuracy
								//	EffectiveArea data server; numerical
								//	problems with polynomial fit from IDL
								//	= _accuracy
	switch (acc) {
		case LOW: {
			CcfConstituentTable ccftab(set, "FIT_PARAM");
			const Table *table = ccftab.block();
			const ColumnData *cdE = table->column("ENERGY")->data(),
							 *cdCoef = table->column("COEF_MAPS")->data();
			real32 *E = cdE->real32Data();
			real64 *coef = cdCoef->real64Data();

			unsigned N = table->rows();
			vector<double> eGrid(N);
			vector<DImage> coefMaps;
			vector<unsigned long> exts = cdCoef->size();
			for (unsigned n=0; n<N; ++n) {
				eGrid[n] = E[n];
				coefMaps.push_back(DImage(exts[0], exts[1]));
				double *map = coefMaps[n].data();
				for (unsigned i=0; i<exts[0]*exts[1]; ++i)
					map[i] = coef[cdCoef->position(n, i)];
			}
			delete cdE;
			delete cdCoef;

			mds = new LowAccuracyEffectiveAreaDataServer(
							ae,
						    eGrid,
						    coefMaps,
							table->attribute("DELTA_F")->asReal(),
							table->attribute("DELTA_Y")->asReal(),
							table->attribute("DELTA_Z")->asReal(),
							this);

			break;
			}
		case HIGH:
			errstr<<"High-accuracy version of EffectiveArea effective area is "
					"not available - reverting to normal accuracy"
				  <<Sas::warning("NotAvailable");
		case MEDIUM: {
			CcfConstituentTable ccftab(set, "VIGNETTING");
			mds = new MediumAccuracyEffectiveAreaDataServer(ae,
													 ccftab.block(),
													 this);
			break;
			}
		default:
			errstr<<"Unknown accuracy"<<Sas::error("ThisShouldNotHappen");
			break;
	}

	dataSetServer->close(set);

	//
	//	now instantiate right EffectiveArea data server
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			(*_collection)[0] =
				new EmosEffectiveAreaDataServer(
									Proxy<EffectiveAreaDataServer>(mds), this);
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			(*_collection)[0] =			
				new RgsEffectiveAreaDataServer(
									Proxy<EffectiveAreaDataServer>(mds), this);
			break;
		default:
			//	for PN the Ae is the full one
			(*_collection)[0] = mds;
	}
} // CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection()



EffectiveAreaDataServer *CalAtomCollection<EffectiveAreaDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<EffectiveAreaDataServer>::select()
