//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EffectiveAreaDataServer --- the EffectiveAreaDataServer class
//
//	DESCRIPTION:
//		A class 'EffectiveAreaDataServer' is defined which provides access to
//		the mirror and total instrumental effective area curves.
//
//	CHANGE HISTORY:
//		UL, 2000-03-10: 1.0.0: - first version
//		UL, 2000-05-30: 1.1.0: - added new RGS-specific way of Ae construction
//		UL, 2000-06-16: 1.2.0: - fixed compilation problem under gcc 2.95.1
//		UL, 2001-10-24: 1.3.0: - added support for empirical correction
//								 curve in RGS Ae
//
//	AUTHOR:
//		U. Lammers (UL), Fri Mar 10 11:39:13 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbool, C. Erd et al.
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/EffectiveAreaDataServer.h,v 1.8 2002/05/24 10:59:18 gvacanti Exp $
//

#ifndef AE_DATA_SERVER_H
#define AE_DATA_SERVER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "GratingDataServer.h"
#include "Proxy.h"

//
//	---( EffectiveAreaCurve )------------------------------------------------
//
class EffectiveAreaCurve : public Slatec::CubicHermiteInterpolant<double>{
public:
	EffectiveAreaCurve() : _theta(0.), _phi(0.) {}
	EffectiveAreaCurve(const std::vector<double> &eGrid,
					   const std::vector<double> &area,
					   double theta = 0., double phi = 0.)
		: Slatec::CubicHermiteInterpolant<double>(eGrid, area),
			_theta(theta), _phi(phi) {}

	EffectiveAreaCurve(const Slatec::CubicHermiteInterpolant<double> &ae,
					   double theta = 0., double phi = 0.)
		: Slatec::CubicHermiteInterpolant<double>(ae),
		  _theta(theta), _phi(phi) {}


	virtual double area(double E) const;
		//
		//	PURPOSE:
		//		Evaluate effective area curve at a single energy <E>
		//
		//	ARGUMENTS:
		//		E	: energy in eV
		//
		//	RETURNS:
		//		Value of effective area curve at <E> in cm^2
		//


	virtual double area(double Elo, double Ehi) const;
		//
		//	PURPOSE:
		//		Return weighted integral of effective area curve between
		//		the	two energies
		//
		//	ARGUMENTS:
		//		Elo, Ehi	: energies in eV between which filter transmission
		//					  is to be integrated; Elo < Elo
		//
		//	RETURNS:
		//						 Ehi
		//                 1    /
		//			F = ------- |   dE Ae(E)
		//              Ehi-Elo /
		//						Elo
		//

	double theta() const { return _theta; }
	double phi() const { return _phi; }
		//
		//	PURPOSE:
		//		Return FOV angles for which this effective area curve is
		//		valid
		//


	virtual ~EffectiveAreaCurve() {}

private:
	double	_theta,	//	off-axis angle [rad]
			_phi;	//	azimuth angle [rad]
};


typedef Slatec::CubicHermiteInterpolant<double> VignettingCurve;


//
//	---( EffectiveAreaDataServer )---------------------------------------------
//
class EffectiveAreaDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<EffectiveAreaDataServer> > {
public:
	EffectiveAreaDataServer(CalAtomCollection<EffectiveAreaDataServer> *parent
							= 0);
		//
		//	PURPOSE:
		//		Instantiate EffectiveAreaDataServer object
		//


	virtual void dump(std::ostream &);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { return Ccf::XAreaEf; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual EffectiveAreaCurve effectiveAreaCurveFull(double theta = 0.,
													  double phi = 0.) const;
	virtual EffectiveAreaCurve effectiveAreaCurve(double theta = 0.,
												  double phi = 0.) const;
	virtual EffectiveAreaCurve effectiveAreaCurve(int spectralOrder,
												  double theta = 0.,
												  double phi = 0.) const;
	virtual EffectiveAreaCurve intrinsicEffectiveAreaCurve(int spectralOrder,
												  double theta = 0.,
												  double phi = 0.) const;
		//
		//	PURPOSE:
		//		Return effective area curve (mirror part only) applicable
		//		at a requested off-axis/azimuth	angle theta/phi
		//		The 'Full' form returns the area curve without the
		//		RGA-vignetting effect (if applicable)
		//
		//		The second last forms are RGS-specific; they also require a
		//		spectral order to be specified (see below)
		//
		//	ARGUMENTS:
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] for which
		//					  effective area curve is requested
		//		spectralOrder: spectral order (RGS)
		//

	virtual double effectiveArea(double E,
								 double theta = 0.,
								 double phi = 0.) const;
		//
		//	RETURNS:
		//		effectiveAreaCurve(theta, phi) evaluated at energy E [eV].
		//


	virtual VignettingCurve vignettingCurve(double theta = 0.,
											double phi = 0.) const;
	virtual VignettingCurve vignettingCurve(int spectralOrder,
											double theta = 0.,
											double phi = 0.) const;
		//
		//	PURPOSE:
		//		Return vignetting curve applicable at a requested
		//		off-axis/azimuth angle theta/phi; the second form is
		//		RGS-specific in that it requires the specification of a
		//		spectral order
		//
		//	ARGUMENTS:
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] for which
		//					  vignetting curve is requested
		//		spectralOrder: spectral order (RGS)
		//


	virtual ~EffectiveAreaDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//
}; // EffectiveAreaDataServer


class LowAccuracyEffectiveAreaDataServer :
									public virtual EffectiveAreaDataServer {
public:
	LowAccuracyEffectiveAreaDataServer(Proxy<EffectiveAreaCurve>
									   onAxisEffectiveAreaCurve,
									   const std::vector<double> &eGrid,
									   const std::vector<DImage> &coefs,
									   double deltaF = 0.,
									   double deltaY = 0.,
									   double deltaZ = 0.,
					CalAtomCollection<EffectiveAreaDataServer> *parent = 0);

	EffectiveAreaCurve effectiveAreaCurve(double theta = 0.,
										  double phi = 0.) const ;
	double effectiveArea(double E, double theta = 0., double phi = 0.) const;

	~LowAccuracyEffectiveAreaDataServer();

protected:
	Proxy<EffectiveAreaCurve>	_onAxisEffectiveAreaCurve;
	std::vector<double>	_eGrid;
	Image<Slatec::CubicHermiteInterpolant<double> *>	_coefsCurve;
	double			_deltaF,
					_deltaY,
					_deltaZ,
					_focalLength;

private:
	pair<double, double> yz(double, double) const;
	double vignettingFactor(double, const pair<double, double> &,
							unsigned, unsigned) const;
}; // LowAccuracyEffectiveAreaDataServer


class MediumAccuracyEffectiveAreaDataServer :
									public virtual EffectiveAreaDataServer {
public:
	MediumAccuracyEffectiveAreaDataServer(Proxy<EffectiveAreaCurve>
										  onAxisEffectiveAreaCurve,
										  const Table *vignettingTable,
					CalAtomCollection<EffectiveAreaDataServer> *parent = 0);

	EffectiveAreaCurve effectiveAreaCurve(double theta = 0.,
										  double phi = 0.) const;
	double effectiveArea(double E, double theta = 0., double phi = 0.) const;

	~MediumAccuracyEffectiveAreaDataServer();

protected:
	Proxy<EffectiveAreaCurve>	_onAxisEffectiveAreaCurve;
	std::vector<double>			_eGrid;
	std::vector<Slatec::CubicHermiteInterpolant<double> *>	_vignetting;

private:
	Slatec::CubicHermiteInterpolant<double>	*_vignettingIndexLocator;
	unsigned vignettingIndex(double) const;
}; // MediumAccuracyEffectiveAreaDataServer


class EmosEffectiveAreaDataServer : public virtual EffectiveAreaDataServer {
public:
	EmosEffectiveAreaDataServer(Proxy<EffectiveAreaDataServer> mds,
					CalAtomCollection<EffectiveAreaDataServer> *parent = 0);

	virtual EffectiveAreaCurve effectiveAreaCurveFull(double theta = 0.,
													  double phi = 0.) const;
	virtual EffectiveAreaCurve effectiveAreaCurve(double theta = 0.,
												  double phi = 0.) const;
	double effectiveArea(double E, double theta = 0., double phi = 0.) const;

	virtual ~EmosEffectiveAreaDataServer() {}

private:
	Proxy<EffectiveAreaDataServer>		_unvignetted;
	GratingDataServer					*_gds;
};


class RgsEffectiveAreaDataServer : public virtual EffectiveAreaDataServer {
public:
	RgsEffectiveAreaDataServer(Proxy<EffectiveAreaDataServer> mds,
					CalAtomCollection<EffectiveAreaDataServer> *parent = 0);

	virtual EffectiveAreaCurve effectiveAreaCurveFull(double theta = 0.,
													  double phi = 0.) const;
		//
		//	PURPOSE:
		//		Compute the effective area curve of the mirror module only
		//		(no RGA intercept effects)
		//
		//	ARGUMENTS:
		//		see above
		//

	virtual EffectiveAreaCurve effectiveAreaCurve(int spectralOrder,
												  double theta = 0.,
												  double phi = 0.) const;
		//
		//	PURPOSE:
		//		Compute the effective area of the RGS according to [2]
		//		as a product of
		//			o Ae of respective mirror module
		//			o RGA intercept efficiency
		//			o RGA reflection efficiency
		//			o RGA self-vignetting efficieny
		//
		//	ARGUMENTS:
		//		see above
		//

	virtual EffectiveAreaCurve intrinsicEffectiveAreaCurve(int spectralOrder,
												  double theta = 0.,
												  double phi = 0.) const;
		//
		//	PURPOSE:
		//		Compute the effective area of RGS as a mere product of
		//			o Ae of respective mirror module
		//			o RGA intercept efficiency
		//			o RGA reflection efficiency
		//		This form is to be used for the response matrix generation
		//		where the inclusion of the RGA self-vignetting efficiency
		//		is conceptually inappropriate (see [2] for details)
		//
		//	ARGUMENTS:
		//		see above
		//


	virtual ~RgsEffectiveAreaDataServer() {}

private:
	Proxy<EffectiveAreaDataServer>		_unvignetted;
	Proxy<GratingDataServer>			_gratingDataServer;	

	std::vector<double>					_empiricalCorrectionEgrid,
										_empiricalCorrectionValue;
										//	multiplicative empirical correction
										//	curve to tweak shape of overall Ae
										//	curve; driven by table
										//	RGS?_QUANTUMEF.CCF:RGA_EFFAREACORR
}; 

#endif
