//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	PatternLibServer	--- the PatternLib class
//
//	DESCRIPTION:
//		Class PatternLibServer provides access to the event pattern library.
//
//	CHANGE HISTORY:
//		UL, 2000-05-17: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Thu May 18 10:10:34 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/PatternLibServer.cc,v 1.10 2002/07/26 15:07:39 ulammers Exp $
//

#include "PatternLibServer.h"

using namespace std;

EventPattern::EventPattern(unsigned short id,
						   unsigned xExtent, unsigned yExtent,
						   const int8 *vector)
	: Image<int8>(xExtent, yExtent, vector), _id(id), _grade(0)
{
	//	compute pattern grade
	unsigned N = size();
	int8 *d = this->data();
	for (unsigned n=0; n<N; ++n)
		if (UNT(d[n] == 1))	//	pixels above threshold are coded as 1
			++_grade;
} // EventPattern::EventPattern()


PatternLibServer::PatternLibServer(const Table *table,
								 CalAtomCollection<PatternLibServer> *parent)
	: CalChild<CalAtomCollection<PatternLibServer> >(parent)
{
	ColumnData *cdpd = table->column("PATTERN")->data(),
			   *cdpattid = table->column("PATTERN_ID")->data();
	int8 *pattdat = cdpd->int8Data(),
		 *pattid = cdpattid->int8Data();

	//	check if the size of the maps is ok
	unsigned long xs = cdpd->size()[0],
			      ys = cdpd->size()[1];
	for (unsigned n=0; n<table->rows(); ++n) {
		PatternLibServer::MapType::const_iterator
											i = _patternMap.find(pattid[n]);

		if (i != _patternMap.end()) {
			//	that pattern exists already - delete it
			errstr<<"Pattern with id "<<INT(pattid[n])<<" exists already "
					"- will overwrite it"
					  <<Sas::warning("OverwritingExistingData");
			delete (*i).second;
		}
		_patternMap[pattid[n]] = new EventPattern(pattid[n], xs, ys,
											  pattdat+cdpd->position(n, 0));
	}

	//	tidy up
	delete cdpd;
	delete cdpattid;
} // PatternLibServer::PatternLibServer()


const EventPattern &PatternLibServer::pattern(unsigned patternId) const
{
	PatternLibServer::MapType::const_iterator i = _patternMap.find(patternId);

	if (i == _patternMap.end())
		errstr<<"There is no pattern with ID "<<patternId<<" defined"
			  <<Sas::error("NoSuchPattern");

	return *(*i).second;
} // PatternLibServer::pattern()


vector<vector<unsigned short> > PatternLibServer::gradePatternList() const
{
	vector<vector<unsigned short> > v;
	for (PatternLibServer::MapType::const_iterator i = _patternMap.begin();
		 i != _patternMap.end(); ++i) {
		const EventPattern &ep = *(*i).second;
		unsigned short grade = ep.grade();
		if (grade > v.size())
			v.resize(grade);
		v[grade-1].push_back(ep.id());
	}

	return v;
} // PatternLibServer::gradePatternList()


PatternLibServer::~PatternLibServer()
{
	for (PatternLibServer::MapType::const_iterator i = _patternMap.begin();
		 i != _patternMap.end(); delete (*i++).second);
} // PatternLibServer::~PatternLibServer()


CalAtomCollection<PatternLibServer>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
	_nEntries	= _nModes;
	PatternLibServer *null = 0;
	_collection = new vector<PatternLibServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::PatternLib);
	DataSet *dataSet = e->open();
	const vector<Mode> &modes = instrument->modes();
	for (unsigned mode=0; mode<instrument->nModes(); ++mode) {
		string blockName = Xmm::toString(modes[mode]);
		if (!e->hasBlock(blockName))
			//	no pattern lib for that mode
			continue;

		CcfConstituentTable ccftab(dataSet, blockName);
		(*_collection)[mode] = new PatternLibServer(ccftab.block(),
												    this);
	} // for(mode=0; ...

	dataSetServer->close(dataSet);
} // CalAtomCollection<PatternLibServer>::CalAtomCollection()


PatternLibServer *CalAtomCollection<PatternLibServer>::select(
					Ccd, Node,
					unsigned modeOffset) const
{
	if (modeOffset > xmmPayLoad.emos1.nModes() || !(*_collection)[modeOffset])
		//	not found
		errstr<<"There is no pattern lib for EMOS mode `"
			  <<Xmm::toString(xmmPayLoad.emos1.modes()[modeOffset])<<"'"
			  <<Sas::error("IllegalInput");

	return (*_collection)[modeOffset];
} // PatternLibServerCollection::select

