//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	PsfDataServer --- the PsfDataServer class
//
//	DESCRIPTION:
//		A class 'PsfDataServer' is defined which provides access to the
//		mirror Points-Spread-Functions and related quantities.
//
//	CHANGE HISTORY:
//		UL, 2000-03-10: 1.0.0: - first version
//		UL, 2000-05-30: 1.1.0: - added new RGS-specific way of Ae construction
//		UL, 2000-06-16: 1.2.0: - fixed compilation problem under gcc 2.95.1
//		UL, 2000-10-17: 1.3.0: - added yet another flavor of
//									XmmPsf::enboxedEnergy()
//							   - all enboxed/encircledEnergy() variants now
//								 make use of origin set with preceding call
//								 to atOrigin()
//		UL, 2000-11-22: 1.4.0: - added dispersionFigureDistribution()
//		UL, 2000-12-20: 1.4.1: - more precise description of
//									dispersionFigureDistribution()
//		UL, 2001-11-01: 1.5.0: - added class CnrAnalyticPsf
//		JDR,2002-06-13: 1.6.0: - additional description of distribution
//								 returned by dispersionFigureDistribution()
//
//	AUTHOR:
//		U. Lammers (UL), Fri Mar 10 11:39:13 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/PsfDataServer.h,v 1.16 2002/07/22 16:04:22 jbakker Exp $
//

#ifndef PSF_DATA_SERVER_H
#define PSF_DATA_SERVER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Proxy.h"

//
//	---( PSF )---------------------------------------------------------------
//
class Psf {
public:
	typedef BinSize<double> BinSize;
	typedef Interval<double> ValidityInterval;

	Psf();
		//
		//	PURPOSE:
		//		Construct generic Point-Spread-Function.


	virtual const Psf &setParameters(double E,
									 double theta = 0.,
									 double phi = 0.);
		//
		//	PURPOSE:
		//		Set characteristic parameters for the PSF. Any attempt
		//		to set the parameters outside the validity limits
		//		reported by eValidity()/thetaValidity/phiValidity()
		//		(see below) will either lead to a mere warning message or
		//		possibly to an error condition. Please note: Evaluating the PSF
		//		outside its reported validity limits are very likely
		//		to yield  largely inaccurate results!
		//
		//	ARGUMENTS:
		//		E, theta, phi: energy [eV], off-axis angle [arcsec], azimuth
		//					   angle [rad] this PSF is valid at - (theta/phi)
		//					   are given in the TELCOORD refernce system
		//


	double E() const { return _E; }
	double theta() const { return _theta; } // * RAD_TO_DEG * 3600.; }
	double phi() const { return _phi; }
		//
		//	PURPOSE:
		//		Return characteristic parameters this PSF is valid for.
		//		E [eV], theta [arcsec], phi [rad]
		// 


	virtual double operator()(double x, double y) const = 0;
		//
		//	PURPOSE:
		//		Return point value of the PSF at a position (x, y) [mm]
		//		in the CAMCOORD2 system shifted to the PSF barycenter.
		//		Hence,
		//			psf(0., 0.)
		//		is the value of the PSF at its barycenter. Total PSF
		//		is normalized to 1:
		//		    /Inf /Inf
		//			|    |	dx*dy PSF(x, y) = 1
		//		   /-Inf /-Inf
		//


	virtual BinSize binSize() const;
	virtual bool binSizeChangeable() const { return true; }
	virtual const Psf &binSize(const BinSize &binSize) const;
		//
		//	PURPOSE:
		//		Inquire/set intrinsic spatial bin size with which the
		//		PSF is to be sampled by below methods; not all implementations
		//		shall allow to change the bin size from its intrinsic
		//		default value. In this case binSizeChangeable() shall return
		//		false and and an attempt to change it will lead to an
		//		error condition.
		//


	virtual ValidityInterval eValidity() const;
	virtual ValidityInterval thetaValidity() const;
	virtual ValidityInterval phiValidity() const;
		//
		//	PURPOSE:
		//		Inquire the validity interval in energy [eV], off-axis [arcsec]
		//		and azimuth [rad] field angle (in arcsec) for which the PSF is
		//		applicable.
		//


	virtual const Psf &atOrigin(double x = 0., double y = 0.) const;
		//
		//	PURPOSE:
		//		Define the spatial origin of the map that is to be returned
		//		by image(); (0/0) refers to the PSF barycenter
		//


	typedef SpatialWindow<double> PsfWindow;
	virtual Image<float> image(const PsfWindow &window) const = 0;
		//
		//	PURPOSE:
		//		Construct an image of the PSF in a specified window
		//
		//	ARGUMENTS:
		//		window:	PSF window inside which PSF image is to be
		//				constructed; dimensions in mm [CAMCOORD2]
		//


	virtual PsfWindow enclosingBox(double fraction,
								   double aspectRatioYoverX = 1.) const = 0;
	virtual double enclosingCircle(double fraction) const = 0;
		//
		//	PURPOSE:
		//		Compute the spatial extents of a rectangular box (of given
		//		fixed aspect ratio) that encloses a given fraction of the
		//		integrated Point-Spread-Function. The second form
		//		computes the radius of a circle that contains the desired
		//		fraction. Makes use of origin set with preceding call to
		//		atOrigin().
		//
		//	ARGUMENTS:
		//		fraction:		PSF fraction that is to lie within the returned
		//					    spatial (rectangular) window
		//		aspectRatioYoverX:
		//						desired height/width ratio of the PSF box to
		//						be used to compute encircled energy; a value
		//						of 1. selects a perfect square;
		//
		//	RETURNS:
		//		PsfWindow extents [mm] / circle radius [mm]
		//


	virtual double enboxedEnergy(const PsfWindow &psfWindow) const = 0;
	virtual double encircledEnergy(double radius) const = 0;
		//
		//	PURPOSE:
		//		Inverse of enclosingBox()/enclosingCircle() above: Compute
		//		the fraction of the PSF that is contained in a given
		//		rectangular window/circle. If the window/circle is infinitely
		//		large the function returns by definition 1. Makes use of
		//		origin set with preceding call to atOrigin().
		//
		//	ARGUMENTS:
		//		psfWindow	: window inside which the PSF is to be integrated
		//		radius		: radius [mm] of circle around PSF
		//
		//	RETURNS:
		//		PSF fraction in given figure [unitless]
		//


	virtual ~Psf() {}
		//
		//	PURPOSE:
		//		Destroy PSF object constructed with above constructor(s)
		//

protected:
	mutable double	_xOffset, _yOffset;
	void eValidity(ValidityInterval eVal) { _eVal = eVal; }
	void thetaValidity(ValidityInterval thetaVal) { _thetaVal = thetaVal; }
	void phiValidity(ValidityInterval phiVal) { _phiVal = phiVal; }

private:
	double	_E, _theta, _phi;
	mutable	BinSize	_binSize;
	mutable	ValidityInterval	_eVal, _thetaVal, _phiVal;
};


class XmmPsf : public Psf {
public:
	XmmPsf(const Table *alignmentParamTable,
		   const Block *binningParameterBlock);

	virtual const Psf &setParameters(double E,
									 double theta = 0., double phi = 0.);

	virtual Image<float> image(const PsfWindow &window) const
	{
		return integratedPsfImage(window).first;
	}

	virtual double enboxedEnergy(const PsfWindow &window) const;

	virtual double encircledEnergy(double radius) const;

	virtual PsfWindow enclosingBox(double fraction,
									double aspectRatioYoverX = 1.) const;
	virtual double enclosingCircle(double fraction) const;

	virtual double operator()(double x, double y) const
	{
		return pointValue(xPsf(x, y), yPsf(x, y));
	}

	virtual double pointValue(double x, double y) const = 0;

	virtual ~XmmPsf() {}

protected:
	double px() const { return _px; }
	double py() const { return _py; }

private:
	typedef std::pair<Image<float>, double>	PsfPair;
	PsfPair integratedPsfImage(const PsfWindow &window) const;
	FocalPlanePosition barycenter() const;
	double xPsf(double xCc2, double yCc2) const;
	double yPsf(double xCc2, double yCc2) const;
	double xCc2(double xPsf, double yPsf) const;
	double yCc2(double xPsf, double yPsf) const;
	double intPsf(int ix, int iy) const;

	InterpolatedCurve	_deltaYInter,
						_deltaZInter,
						_deltaFInter;
	double				_F;

	mutable double		_px, _py,				//	intrinsic bin size
						_oneOverPx,	_oneOverPy,	//	1/intrinsic bin size
						_nx, _ny,				//	extents in X/Y
						_baryX, _baryY,			//	PSF barycenter in CC2
						_deltaX,
						_deltaY,				//	correction factors
						_Fbar,					//	(F+_deltaF)*tan(theta)
						_sinPhi,				//	sin(phi)
						_cosPhi;				//	cos(phi)

	mutable double		_binX, _binY,
						_binArea;
};


//
//	Interpolated2DParamDataServer: auxiliary class to perform interpolation
//								   interpolation in two dimensions on a
//								   variable number of parameters
//
class Interpolated2DParamDataServer {
public:
	Interpolated2DParamDataServer(const Table *fitParamTable,
								  const std::string &p1ColName,
								  const std::string &p2ColName,
								  const std::string &paramColName);

	unsigned parameters() const { return _nParams; }
	double operator()(unsigned n, double p1, double p2) const {
		if (n>=_nParams)
			errstr<<"Requested value of parameter "<<n<<" but there are only "
				  <<_nParams<<Sas::error("TooFewParameters");

		return (*_interpolatedSurfaces[n])(p1, p2);
	}

	std::vector<double> operator()(double p1, double p2) const {
		std::vector<double> v(_nParams);
		for (unsigned n=0; n<_nParams; ++n)
			v[n] = (*_interpolatedSurfaces[n])(p1, p2);

		return v;
	}

	~Interpolated2DParamDataServer();

private:
	unsigned _nParams;	//	number of served parameters
	std::vector<InterpolatedSurface *>	_interpolatedSurfaces;
};


class CslAnalyticPsfLow : public XmmPsf {
public:
	CslAnalyticPsfLow(const Table *alignmentParamTable,
					  const Table *fitParamTable);

	virtual const Psf &setParameters(double E,
									 double theta = 0.,
									 double phi = 0.);

	double pointValue(double x, double y) const;

	~CslAnalyticPsfLow();

private:
	Interpolated2DParamDataServer		_paramDataServer;
	std::vector<double>					_param;
	double								_norm;
}; // CslAnalyticPsfLow


class CslAnalyticPsfMedium : public XmmPsf {
public:
	CslAnalyticPsfMedium(const Table *alignmentParamTable,
						 const Table *fitParamTable);

	virtual const Psf &setParameters(double E,
									 double theta = 0.,
									 double phi = 0.);

	double pointValue(double x, double y) const;

	~CslAnalyticPsfMedium();

private:
	std::vector<double>				_Ap, _Bp, _Cp;
	double							_norm;
}; // CslAnalyticPsfMedium


class SmallMapsBasedPsf : public XmmPsf {
public:
	SmallMapsBasedPsf(const Table *alignmentParamTable,
					  const CifEntry *cifEntry,
					  const Array *firstSmallMap);

	bool binSizeChangeable() const { return false; }

	virtual const Psf &setParameters(double E,
									 double theta = 0.,
									 double phi = 0.);

	double pointValue(double x, double y) const;

	~SmallMapsBasedPsf();

private:
	DataSet												*_psfSet;
	typedef std::map<double, unsigned, std::less<double> > MapTypeInner;
	typedef MapTypeInner::iterator						MIterInner;
	typedef std::map<double, MapTypeInner, std::less<double> > MapTypeOuter;
	typedef MapTypeOuter::iterator						MIterOuter;
	MapTypeOuter										_idMap;
	//	Energy [eV]	  theta [']	ArrayId

	class PsfMap : public Image<float> {
	public:
		PsfMap(Array *a);
		unsigned id() const { return _id; }
		double E() const { return _E; }
		double theta() const { return _theta; }

		~PsfMap() {}
	private:
		unsigned	_id;
		double		_E, _theta;
	};

	PsfMap	*_maps[4];	// LL, LR, UR, UL
	double		_tE, _ut;
	int			_xExtent, _yExtent;
}; // SmallMapsBasedPsf


class CnrAnalyticPsf : public XmmPsf {
public:
	CnrAnalyticPsf(const Table *alignmentParamTable,
				   const Table *fitParamTable);

	virtual const Psf &setParameters(double E,
									 double theta = 0.,
									 double phi = 0.);

	double enclosingCircle(double fraction) const;	//	provide special version
	double encircledEnergy(double radius) const;	//	provide special version

	double pointValue(double x, double y) const;

	~CnrAnalyticPsf();

private:
	Interpolated2DParamDataServer	_paramDataServer;
	std::vector<double>				_param;	//	0: rc  1: alpha
	double							_norm,	//	normalization factor []
									_rc2,	//	square of core radius [arcmin]
									_alpha1,//	alpha-1 []
									_fac,	//	M_PI*_rc2/_alpha1
									_normAt;//	normalization radius [arcmin]
}; // CnrAnalyticPsf


//
//	---( PsfDataServer )------------------------------------------------
//
class PsfDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<PsfDataServer> > {
public:
	PsfDataServer(Psf *psf,
				  const Table *betaPsfTab,
				  CalAtomCollection<PsfDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate PsfDataServer object
		//


	virtual void dump(std::ostream &);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { return Ccf::XPSF; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual Psf::ValidityInterval eValidity() const;
	virtual Psf::ValidityInterval thetaValidity() const;
	virtual Psf::ValidityInterval phiValidity() const;
		//
		//	PURPOSE:
		//		Inquire energy/theta/phi validity limits of this PSF
		//


	virtual Psf &psf(double E, double theta = 0., double phi = 0.);
		//
		//	PURPOSE:
		//		Return reference to PSF instance applicable at given
		//		input parameters; requesting PSF outside its validity
		//		limits might lead to an error condition
		//
		//	ARGUMENTS:
		//		E		: energy [eV]
		//		theta	: off-axis field angle (TELCOORD) [arcsec]
		//		phi		: azimuth field angle (TELCOORD) [rad]
		//


	virtual Spectrum dispersionFigureDistribution(
											  double nbetaHWHM,
											  double betaChannelWidth,
											  double chromaticMagnification,
											  double inChannelOffset = 0.,
											  bool analytical = true);
		//
		//	PURPOSE:
		//		Compute the figure of the mirror Point-Spread-Function
		//		along the RGA dispersion direction as a one-dimensional
		//		probability distribution. See below for more details
		//		on the correct interpretation of the returned array.
		//
		//	ARGUMENTS:
		//		nBetaHWHM:		  number of Half-Width-Half-Maximum units to
		//					  evaluate the distribution to; this determines
		//					  the number of channels of the output spectrum
		//		betaChannelWidth:
		//					  width of channels of beta grid [rad]
		//		chromaticMagnification:
		//					  real number that describes the scaling of the
		//					  widths and positions of the distribution's
		//					  components; normally calculated as
		//							sin(alpha)/sin(beta_M)
		//		inChannelOffset:
		//					  offset of the first moment of the underlying
		//					  probability density from the center of the
		//					  returned array; in essence the phase of the
		//					  sampling interval; any offset is permitted,
		//					  but the expected range for this argument is
		//							-.5<=inChannelOffset<+.5
		//		analytical:
		//					  boolean parameter to determine whether
		//					  the distribution is to be constructed using
		//					  a dedicated analytical approximation ('true')
		//					  or by numeric integration of the intrinsic
		//					  mirror PSF in the cross dispersion direction
		//					  ('false')
		//
		//	RETURNS:
		//		The return value is an array F[i] that is to be interpreted
		//		in the following context:
		//
		//		The continuous beta space is discretized as a one dimensional
		//		grid:
		//			beta(x) = betaRef + x * betaChannelWidth
		//		Integer values of x define the beta channel centers.  The
		//		value of betaRef is transparent to the caller; it is not
		//		passed in or out of	the function.  The caller is expected to
		//		align the first moment of the returned array to the appropriate
		//		 reference point.
		//
		//		F[i] (i integer) is the probability that a photon possesses a
		//		beta value in the interval [beta(i-0.5), beta(i+0.5)].
		//		F[i] has been constructed by integrating the underlying
		//		probability density P(beta)	between these limits.
		//
		//		The first moment mu0 of P(beta) is located in F[i] at
		//			i = (int) floor(F.size()/2 + inChannelOffset)
		//		Note that the size of the returned array is always odd and, if
		//		inChannelOffset is kept within the expected range, mu0 is
		//		always located within its center element.
		//

	virtual ~PsfDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

	private:
		Psf				*_psf;
		bool			_hasBetaPsf;
		std::vector<double>	_norm[2],	//	normalization ...
							_center[2],	//	center ...
							_width[2];	//	width of Gaussian/Lorentzian
}; // PsfDataServer

#endif
