//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrector.cc --- implementation of TimeCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-11-23: 1.0.0: - first version
//		UL, 1999-11-29: 1.1.0: - bug fix for treatment of timing mode
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//		[2]: Time resolution capabilities of the XMM EPIC pn-CCD in different
//			 readout modes, M. Kuster et al.
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/cal/src/TimeCorrector.cc,v 1.14 2002/08/23 16:35:36 ulammers Exp $
//
#include "errstr.h"
#include "common.h"
#include "CalOalUtility.h"
#include "CalServer.h"
#include "Xmm.h"

#include "TimeCorrector.h"

using namespace std;

//
//	---( TimeCorrector )---------------------------------------------------
//
vector<double> TimeCorrector::frameTimes() const
{
	return vector<double>();
} // TimeCorrector::frameTimes()


double TimeCorrector::frameTime() const
{
	vector<double> v = frameTimes();
    unsigned modeId = calServer.instrument()->modeIndex(
                                            calServer.state()->mode()->get());
	if (modeId >= v.size())
		errstr<<"Cannot determine frame time for mode index "<<modeId
			  <<Sas::error("UnknownFrameTime");

	return v[modeId];
} // TimeCorrector::frameTime()


const std::vector<std::vector<double> > &
	TimeCorrector::correctionParameters() const
{
	static vector<vector<double> > _v;

	return _v;
} // TimeCorrector::correctionParameters()


CalAtomCollection<TimeCorrector>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nModes		= instrument->nModes();
	_nEntries	= 1;
	TimeCorrector *null = 0;
	_collection = new vector<TimeCorrector *>(_nEntries, null);
	TimeCorrector *&corrector = (*_collection)[0];

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::TimeCorr);
	DataSet *dataSet = e->open();

	//
	//	the rest is instrument specific
	//
	switch (instrument->id()) {
		case Xmm::EMOS1: {
		case Xmm::EMOS2:
			CcfConstituentTable frameTimeTab(dataSet, "FRAMETIME");
			corrector = new EmosTimeCorrector(frameTimeTab, this);
			break;
		}
		case Xmm::EPN: {
			CcfConstituentTable frameTimeTab(dataSet, "FRAMETIME"),
								timeCorrTab(dataSet, "TIMECORR");
			corrector = new EpnTimeCorrector(frameTimeTab, timeCorrTab, this);
			break;
		}
		default:
			corrector = new TimeCorrector(this);
			break;
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<TimeCorrector>::CalAtomCollection()


void TimeCorrector::dump(ostream &os)
{
	os<<"generic TimeCorrector{}";
} // TimeCorrector::dump()


TimeCorrector *CalAtomCollection<TimeCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	_ccd = ccd;
	_node = node;

	return (*_collection)[0];
} // CalAtomCollection<TimeCorrector>::select



void TimeCorrector::noOp() const
{
//	errstr<<"Generic time corrector cannot perform fine time adjustments - "
//			"leaving input unchanged"
//		  <<Sas::warning("NoOperation");
} // TimeCorrector::noOp()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time) const
{
	noOp();

	return time;
} // TimeCorrector::correct()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time,
										CalInt16Vector &) const
{
	noOp();

	return time;
} // TimeCorrector::correct()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time,
										CalInt16Vector &, int16) const
{
	noOp();

	return time;
} // TimeCorrector::correct()


//
//	-------------( EpnTimeCorrector )----------------------------------------
//
EpnTimeCorrector::EpnTimeCorrector(const CcfConstituentTable &frameTimeTab,
								   const CcfConstituentTable &timeCorrTab,
								   CalAtomCollection<TimeCorrector> *parent)
		: CalChild<CalAtomCollection<TimeCorrector> >(parent),
		  TimeCorrector(parent)
{
	//	read frame times
	const Table *table = frameTimeTab.block();
	vector<unsigned short> modeId;
	CalOalUtility::fillVector(table, "MODE_ID", modeId);
	if (!CalCommon::sorted(modeId))
		errstr<<"`"<<table->qualifiedName()<<"' is not sorted in ascending "
				"order of MODE_ID"<<Sas::error("UnsortedModeId");

	vector<vector<double> > v;
	CalOalUtility::fillMatrix(table, "FRAME_TIME", v);
	assert(v.size()); _frameTimes = v[0];

	//	read correction parameters
	table = timeCorrTab.block();
	vector<unsigned short> modeId2;
	CalOalUtility::fillVector(table, "MODE_ID", modeId2);
	if (!CalCommon::sorted(modeId2))
		errstr<<"`"<<table->qualifiedName()<<"' is not sorted in ascending "
				"order of MODE_ID"<<Sas::error("UnsortedModeId");

	CalOalUtility::fillMatrix(table, "FRAME_TIME", _coefs);
	if (_coefs.size() != 3)
		errstr<<"Expected to see three elements in column `"
			  <<table->qualifiedName()<<":FRAME_TIME' but there are "
			  <<_coefs.size()<<Sas::error("DataMismatch");
} // EpnTimeCorrector::EpnTimeCorrector()


CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time) const
{
	//	the following implements Equations (1)-(6) + (10) in Ref. [2]

	//	retrieve mode / current CCD number and break latter in quadrant and
	//	ccd id [0, 2]
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	Ccd ccd = calServer.state()->ccd()->get();
	unsigned ccdId = EPNcam::toCcdId(ccd);

	double deltaT = 0.;

	switch (mode) {
		case Xmm::PRIME_FULL_WINDOW:
		case Xmm::PRIME_FULL_WINDOW_EXTENDED:
		case Xmm::PRIME_LARGE_WINDOW:
			deltaT = _coefs[ccdId][modeId];
			break;

		case Xmm::PRIME_SMALL_WINDOW:
			if (ccd != 4)
				errstr<<"In `"<<mode<<"' mode only CCD #4 can be active but "
						"the current CCD is "<<ccd
					  <<Sas::error("IllegaCcd");
			deltaT = _coefs[0][modeId];
			break;

		case Xmm::FAST_BURST:
			deltaT = _coefs[0][modeId];
			break;

		default:
			errstr<<"Scalar fine time adjustment in mode `"<<mode<<"' it not "
					"supported"<<Sas::error("UnsupportedOperation");
			break;
	}

	//	convert deltaT from ms to s
	deltaT *= 1.e-3;

	unsigned N = time.size();
	for (unsigned n=0; n<N; ++n)
		time[n] -= deltaT;	//	Eq. (1)-(6) + (10) in [2]

	return time;
} // EpnTimeCorrector::correct()


CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time,
										   CalInt16Vector &rawy) const
{
	//	the following implements Eq. (7) in Ref. [2]

	//	get mode from Cal state
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	if (mode != Xmm::FAST_TIMING)
		errstr<<"Fine-time adjustment with RAWY and no source position in "
				"mode `"<<mode<<"' is not supported - this is only possible "
				"in mode"<<Xmm::FAST_TIMING
			  <<Sas::error("UnsupportedOperation");

	double p1 = _coefs[0][modeId] * 1.e-3,
		   p2 = _coefs[1][modeId] * 1.e-3;

	unsigned N = time.size();
	for (unsigned n=0; n<N; ++n)
		time[n] -= (9-INT(DBL(rawy[n]-1)/20.)) * p1 - p2;	//	Eq. (7) in [2]

	return time;
} // EpnTimeCorrector::correct()


CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time,
										   CalInt16Vector &rawy,
										   int16 sourceY) const
{
	//	the following implements Eq. (9) + (11) in Ref. [2]

	//	get mode from Cal state
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	unsigned N = time.size();
	if (mode == Xmm::FAST_TIMING) {
		double p = _coefs[2][modeId] * 1.e-3;

		for (unsigned n=0; n<N; ++n)
			time[n] += (rawy[n]-1 - 200 - INT(.1*DBL(sourceY-1))) * p;
														//	Eq. (9) in [2]
	} else if (mode == Xmm::FAST_BURST) {
		double p = _coefs[1][modeId] * 1.e-3,
			   a = _coefs[2][modeId] * 1.e-3;

		for (unsigned n=0; n<N; ++n)
			time[n] -= a + DBL(sourceY-1 - rawy[n]-1) * p;
														//	Eq. (11) in [2]
	} else
		errstr<<"Fine-time adjustment with source position in mode "<<mode
			  <<" is not supported - must be "<<Xmm::FAST_TIMING
			  <<" or "<<Xmm::FAST_BURST
			  <<Sas::error("UnsupportedOperation");

	return time;
} // EpnTimeCorrect::correct()


//
//	EMOS
//
EmosTimeCorrector::EmosTimeCorrector(const CcfConstituentTable &frameTimeTab,
									 CalAtomCollection<TimeCorrector> *parent)
		: CalChild<CalAtomCollection<TimeCorrector> >(parent),
		  TimeCorrector(parent)
{
	//	read frame times
	const Table *table = frameTimeTab.block();
	vector<unsigned short> modeId;
	CalOalUtility::fillVector(table, "MODE_ID", modeId);
	if (!CalCommon::sorted(modeId))
		errstr<<"`"<<table->qualifiedName()<<"' is not sorted in ascending "
				"order of MODE_ID"<<Sas::error("UnsortedModeId");

	CalOalUtility::fillMatrix(table, "FRAME_TIME", _frameTimes);
} // EmosTimeCorrector::EmosTimeCorrector()


vector<double> EmosTimeCorrector::frameTimes() const
{
	unsigned ccd = calServer.state()->ccd()->get()-1;
	if (ccd >= _frameTimes.size())
		errstr<<"Cannot determine EMOS frame times for CCD "<<ccd+1
			  <<Sas::error("FrameTimeUnknown");

	return _frameTimes[ccd];
} // EmosTimeCorrector::frameTimes()
